# Stubs for gettext (Python 3.4)

from typing import Any, IO, List, Optional, Union, Callable

class NullTranslations:
    def __init__(self, fp: IO[str] = ...) -> None: ...
    def add_fallback(self, fallback: NullTranslations) -> None: ...
    def gettext(self, message: str) -> str: ...
    def lgettext(self, message: str) -> str: ...
    def ngettext(self, singular: str, plural: str, n: int) -> str: ...
    def lngettext(self, singular: str, plural: str, n: int) -> str: ...
    def info(self) -> Any: ...
    def charset(self) -> Any: ...
    def output_charset(self) -> Any: ...
    def set_output_charset(self, charset: Any) -> None: ...
    def install(self, names: List[str] = ...) -> None: ...

class GNUTranslations(NullTranslations):
    LE_MAGIC: int
    BE_MAGIC: int

def find(domain: str, localedir: str = ..., languages: List[str] = ...,
         all: bool = ...): ...

def translation(domain: str, localedir: str = ..., languages: List[str] = ...,
                class_: Callable[[IO[str]], NullTranslations] = ...,
                fallback: bool = ..., codeset: Any = ...) -> NullTranslations: ...

def install(domain: str, localedir: str = ..., codeset: Any = ...,
            names: List[str] = ...): ...

def textdomain(domain: str = ...) -> str: ...
def bindtextdomain(domain: str, localedir: str = ...) -> str: ...
def bind_textdomain_codeset(domain: str, codeset: str = ...) -> str: ...
def dgettext(domain: str, message: str) -> str: ...
def ldgettext(domain: str, message: str) -> str: ...
def dngettext(domain: str, singular: str, plural: str, n: int) -> str: ...
def ldngettext(domain: str, singular: str, plural: str, n: int) -> str: ...
def gettext(message: str) -> str: ...
def lgettext(message: str) -> str: ...
def ngettext(singular: str, plural: str, n: int) -> str: ...
def lngettext(singular: str, plural: str, n: int) -> str: ...

Catalog = translation



from __future__ import unicode_literals

from collections import defaultdict

__all__ = (
    'MouseHandlers',
)


class _Rect:
    """
    Internal rectangle class for efficient mouse handler regions.
    """
    def __init__(self, x_min, x_max, y_min, y_max, handler):
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max
        self.handler = handler
    
    def contains(self, x, y):
        """
        Check if the point (x, y) is within this rectangle.
        """
        return (self.x_min <= x < self.x_max and 
                self.y_min <= y < self.y_max)


class RectDict:
    """
    Dictionary-like class that efficiently stores rectangle regions and their handlers.
    """
    def __init__(self, default_factory):
        self._default_factory = default_factory
        self._default_value = default_factory()
        self._rects = []
        
    def __getitem__(self, key):
        """
        Get handler for a specific (x, y) position.
        """
        x, y = key
        
        # Check through rectangles in order (most recently added first)
        for rect in self._rects:
            if rect.contains(x, y):
                return rect.handler
                
        # Return default value if no rectangle contains this point
        return self._default_value
    
    def __setitem__(self, key, value):
        """
        For individual points (legacy support and special cases)
        """
        x, y = key
        # Create a 1x1 rectangle for this point
        rect = _Rect(x, x+1, y, y+1, value)
        self._rects.insert(0, rect)


class MouseHandlers(object):
    """
    Two dimentional raster of callbacks for mouse events.
    """
    def __init__(self):
        def dummy_callback(cli, mouse_event):
            """
            :param mouse_event: `MouseEvent` instance.
            """

        # Map (x,y) tuples to handlers using our optimized RectDict
        self.mouse_handlers = RectDict(lambda: dummy_callback)

    def set_mouse_handler_for_range(self, x_min, x_max, y_min, y_max, handler=None):
        """
        Set mouse handler for a region.
        """
        # Add rectangle directly to the RectDict's internal rect list - O(1) operation
        rect = _Rect(x_min, x_max, y_min, y_max, handler)
        self.mouse_handlers._rects.insert(0, rect)



def CommonPrefix(s1, s2):
	# 0.21143    seconds
    out = ''
    for i, j in zip(s1, s2):
        if i != j:
            break
        out += i
    return out

def slow_longest_common_prefix(seq1, seq2):
	# TICTOC: 0.63711    seconds
    start = 0
    while start < min(len(seq1), len(seq2)):
        if seq1[start] != seq2[start]:
            break
        start += 1
    return seq1[:start]



 >>> longest_common_suffix('12345abcd','876323abcd')
ans = abcd
 >>> longest_common_suffix('adofoieabcd','29348psaabcd')
ans = abcd
 >>> longest_common_suffix('adofoieabcd','29348psaabqcd')
ans = cd
 >>> longest_common_suffix([1,2,3,4,5],[7,6,3,4,5])
ans = [3, 4, 5]
 >>> longest_common_suffix([1,2,3,4,5],[7,3,3,4,3,6,3,4,5])
ans = [3, 4, 5]



#TODO: This file is NOT comlpete yet. TODO: Complete it and make it useable.
#TODO: Make a real test for it, docstring how this is meant to be used for heavy preprocessing like captioning videos, getting tracking ponits, warping noise etc on the fly
#TODO: Make an idle processor: somehow I'd like the dataset to start getting output tickets for 0,1,2,3,4...etc if it's not being queried at a given moment. Might need to modify webevaluator for this
#TODO: In web evaluator, we might want to add a self-repairing option to the clusters so if one hangs and takes too long we can kill it, and it will reboot
#TODO: In web evaluator, we might want to add an auto-ranking option to the delegatees so we can divy GPU's if one crashes
#TODO: Add a light "light_postprocess(input_ticket, output_ticket, sample) function that's meant to be used on the final training dataloader - so that cheap ops that generate large things can be done without having to load from disk (i.e. placing dots on a video - we don't want MP4 compression and its fast enough to calculate, so I don't want to send it over network. The docstrings should specify that this only really matters for the remote dataset options - for local dataset you wont notice a difference between doing it there and putting it directly in the output ticket --> sample function, so beware - if handled right you can make this dataloader sing)
#TODO: IN docs, note that input tickets CAN be any type - not just json-like objects - but if they're large it will be slow over network
#TODO: Add some basic common preprocessing things we might want out of the box into a separate module. For example, an easy way to encode videos with cogx's encoder or caption using SA2VA.
#TODO: A RobustDataset mixin or decorator or something: That when you call __getitem__, and it errors with exception, it chooses a random index and tries again.
#TODO: Note that input tickets could be as simple as a single index, referencing some ticket file + an index! (would have to build a cache for that). Like, we could have each worker have a dataset to modify and play with...
#TODO: If we want to download videos etc, we would want a way to preserve the original downloaded files. Omni save/load should always save an extension if given bytestring always use bytes_to_file. But ideally we'd be able to just download + hardlink or something...
#TODO: In the ticket processor where we pass it lambdas and stuff right now, but let it handle the paths later...but what if they want to use each other? Then we wouldn't be able to do the file caching thing...we probably want to make a decorator INSIDE that function. EDIT: Done!
#TODO: Any JSON's etc, containing extra input ticket info might not be recorded...is this ok? We might want a func to invalidate the cache of some samples...think about it...
import rp
rp.pip_import('dict_hash')

import dict_hash
import rp.web_evaluator as wev
from easydict import EasyDict

class EvaluableEasydict(EasyDict):
    """
    EXAMPLE:
        >>> d=EvaluableEasydict(a=4,b= lambda d:d.a+4)
        >>> d
        {'a': 4, 'b': <function <lambda> at 0x1148c07b8>}
        >>> d.a
        4
        >>> d.b
        8
        >>> d
        {'a': 4, 'b': 8}
    """
    def __getitem__(self, index):
        value = super().__getitem__(index)
        if callable(value):
            value = value(self)
            self[index] = value
        return value
    def __getattribute__(self,attr):
        if attr in self:
            return self[attr]
        return super().__getattribute__(attr)

class TicketHandler(rp.CachedInstances):
    #Input tickets are json-like easydicts
    #Output tickets are easydict dict[str->Any Type, but particularly paths to files]
    #This class might be instantiated many times, so if something needs GPU and need to load a model,
    #best to do that globally as a singleton.
    #KEEP THIS CLASS LEAN! It will be pickled and passed by value at each call. This class should NOT hold much data.
    #No more than a kilobyte preferably.

    def __init__(self, cache_dir):
        self.cache_dir = cache_dir

    def get_output_ticket(self, input_ticket):
        #THIS BLOCK SHOULD BE ABSTRACTED
        #AND DOCSTRING IT...IT SHOULD BE SIMPLE TO UNDERSTAND IF NOT WE NEED TO RETHINK IT
        output_ticket = rp.as_easydict()
        processed_ticket = self.ticket_processor(input_ticket)
        processed_ticket = EvaluableEasydict(processed_ticket)
        sample_folder = self.get_sample_folder(input_ticket)

        #TODO: Turn processed_ticket into a special object where if we access an attr of a callable, it acts like a property and evaluates itself
        #      and also wrap each one of those funcs so that if it evaluates itself, it replaces its value in the original dict by the returned value
        #      this lets them cross-reference each other so for example caption_video could call get_video.
        #UPDATE: Done! We did this todo! See EvaluableEasydict

        for name, value in processed_ticket:
            if name in self.CACHE_PATHS:

                file_name = self.CACHE_PATHS[name]
                file_path = rp.path_join(sample_folder, file_name)

                if not self.path_exists(file_path):
                    #If value is passed as a callable, only evaluate it 
                    #if we need to fill the cache
                    if callable(value):
                        value = value()

                    #Compute the value and cache it
                    self.save_file(value, file_path)
                    output_ticket[value] = file_path

        return output_ticket

    def save_ticket(self, output_ticket, folder=None):
        #output_ticket is dict of filenames -> values
        #For things without an extension, we pass them by value - so we don't save them
        #Returns an output ticket (easydict of str -> value)
    
        output_ticket = rp.as_easydict()

        for name, value in output_ticket.items():
            #Like key=="video.mp4" --> We save the value
            #Key like "video" --> we return the value directly
            
            if name in self.CACHE_PATHS:
                file_path = rp.path_join(folder, self.CACHE_PATHS[name])
                value = self.save_file(value, file_path)
                assert isinstance(value, str), f'self.save_file should return a path, but got type {type(value)}'
            
            output_ticket[name] = value
        
        return output_ticket
    
    def get_ticket_hashname(input_ticket):
        #Maybe you want some semantic info in the name...idk. So this is overrideable
        #But should be unique per input ticket!
        return dict_hash.sha256(input_ticket)[:10]
    
    def get_sample_folder(self, input_ticket):
        return rp.path_join(self.cache_dir, self.get_ticket_hashname(input_ticket))

    def load_sample(self, output_ticket):
        return rp.as_easydict(
            {
                key: self.load_file(value) if isinstance(value, str) and rp.get_file_name(value) in self.CACHE_PATHS else value
                for key, value in output_ticket.items()
            }
        )
    
    def __repr__(self):
        ...
        #TODO
    


    ##### THE BELOW FUNCTIONS ARE COMPUTE-PLATFORM-SPECIFIC #####
    ##### IF CHANGED, IT SHOULD BE VIA A PLATFORM MIXIN OR SOMETHING... #####
    ##### IF YOU HAVE NFS, OR ARE WORKING ON A SINGLE MACHINE - NO CHANGES ARE NEEDED #####

    def save_file(self, object, path):
        #Save a file, return the file path as str
        return rp.r._omni_save(object, path)
    
    def load_file(self, path):
        #On XCloud it would be nice to have used 
        #If you're using cloud storage like AWS, this is the place to put the download functionality!
        #If you ever have to download anything (and not just read the path), make sure you delete it before exiting this function
        return rp.r._omni_load(path)
    
    def path_exists(self, path):
        #NOTE: If you're going to use URL's or such you should implement another thing to check if said URL's exist here, this only checks paths right now...
        return rp.path_exists(path)

    def paths_exist(self, paths):
        #Might be room for optimization on some platforms...
        return all(self.path_exists(path) for path in paths)


    ##### THE BELOW CODE SHOULD BE CHANGED ON A PER-DATASET BASIS #####
    ##### BE IT A VIDEO DATASET, IMAGE DATASET, SNORLAX DATASET #####
    ##### ALL PREPROCESSING IS SPECIFIED IN THE BELOW FUNCTION, GPU OR NOT #####

    #ALL KEYS SHOULD BE NAMESPACEABLE
    #PATHS CAN BE NESTED IF NEEDED, LIKE 'low_quality/video.mp4'
    CACHE_PATHS = dict(
        video_480p49 = 'video.mp4',
        input_ticket = 'input_ticket.json',
    )

    def ticket_processor(self, input_ticket):
        #TODO: THIS FUNCTION NEEDS A BETTER NAME
        #TODO: Docstring this func, explain how the output should be dict of namespaceable names to either callables or values
        #And that if in CACHE_PATHS, it will save it in cache as a file 
        #And that also, if in CACHE_PATHS if given as a callable, will be called and evaluated IF AND ONLY IF we need to fill the cache
        #This is the sauce you need to change for better dataloaders
        #This is a potential example
        #This func does not handle caching! That's handled elsewhere
        #You'll want to subclass this func if you need GPU functionality and preprocessing etc
        #We're providing just the basics here
        #Yugly lol this docstring is messy mess ur job to clean it

        def get_video(processed_ticket):
            raw_video = rp.load_video(input_ticket.raw_video_path)
            video = rp.resize_list(rp.resize_images(raw_video, size=(480,720)),49)
            with rp.temporary_seed_all(input_ticket.seed):
                if rp.random_chance():
                    #Augmentation via random seed example
                    video = video.flip(2)
            return video

        def get_canny_video(processed_ticket):
            #For demo purposes, usually we'd save this for postprocessing and put something much more computationally intense here like captioning or tracking
            video = processed_ticket.video_480p49
            video = [rp.auto_canny(x) for x in video]
            return video
            
        return dict(
            #TODO: Let us choose the ordering using ordereddict or something, so we can choose the order in which these evaluate (if not in cache)
            #Reminder again, if some func here (like canny, which is actually a pretty bad example), is slower to LOAD than it is to CALCULATE, it should not be here.
            #Instead, it should be post-processed
            canny_video = canny_video,
            video_480p49 = get_video,
            input_ticket = input_ticket,
            **input_ticket, #Includes prompt, raw_video_path - passed by value because not in CACHE_PATHS
        )

    def post_process(self, input_ticket, output_ticket, sample):
        #You might want to do something here like inverting the image etc, cheap ops that take a lot of memory and bandwidth to transfer
        return sample
    
class LocalDataset:
    def __init__(self, input_tickets, ticket_handler: TicketHandler):
        self.input_tickets = self.load_input_tickets(input_tickets)
        self.ticket_handler = ticket_handler

    def load_input_tickets(self, input_tickets):
        #Ticket roll can be passed as value or by path
        #Path could be a .csv or a .tsv or an list of [str->[int, float, str]] dicts pickled value
        #If you want to preprocess the list of tickets, playing with seeds etc -
        #  or mix multiple together, this is the place to do it!

        if isinstance(input_tickets, str):
            input_tickets = self.load_file(input_tickets)

        if hasattr(input_tickets, 'to_dict'):
            #Given a pandas table, turn it into a list of dicts
            input_tickets=rp.list_dict_transpose(input_tickets.to_dict())

        #Output should be a list of easydicts, all with string keys
        return input_tickets

    def get_output_ticket(self, input_ticket):
        return self.ticket_handler.get_output_ticket(input_ticket)
    
    def __getitem__(self, index):
        input_ticket = self.input_tickets[index]
        output_ticket = self.get_output_ticket(input_ticket)
        sample=self.load_sample(output_ticket)
        sample=self.ticket_handler.post_process(input_ticket, output_ticket, sample)
        return sample

    def __len__(self):
        return len(self.input_tickets)
    
    def __repr__(self):
        return f'{type(self).__name__}(len={len(self)}, ticket_handler={self.ticket_handler})'
    
class RemoteDataset(LocalDataset):
    def __init__(self, input_tickets, ticket_handler:TicketHandler, delegator: wev.DelegationClient):
        super().__init__(input_tickets=input_tickets, ticket_handler=ticket_handler)
        self.delegator=delegator
    
    def get_output_ticket(self, input_ticket):
        output = self.delegator.evaluate(
            "ticket_handler.get_output_ticket(input_ticket)",
            ticket_handler=self.ticket_handler,
            input_ticket=input_ticket,
        )
        if output.errored:
            raise output.error
        return output.value
        
    def __repr__(self):
        return f'{type(self).__name__}(len={len(self)}, ticket_handler={self.ticket_handler}, delegator={self.delegator})'

    @classmethod
    def launch_cluster(
        cls,
        input_tickets,
        ticket_handler,
        num_workers=8,
        *,
        session_name=None,
        base_port=None,
        python_executable=None,
        command_before=None,
        attach=False,
        roster=None,
        delegator_port=None,
        ip_address=None,
        # Different from defaults
        if_exists="replace",
    ):
        if session_name is None:
            session_name = cls.__name__

        cluster_info = wev.launch_tmux_delegation_cluster(
            num_delegatees=num_workers,
            base_port=base_port,
            python_executable=python_executable,
            command_before=command_before,
            attach=attach,
            roster=roster,
            delegator_port=delegator_port,
            ip_address=ip_address,
            if_exists=if_exists,
        )

        delegator = cluster_info.delegator

        return cls(
            input_tickets=input_tickets,
            ticket_handler=ticket_handler,
            delegator=delegator,
        )




from typing import IO, Any, AnyStr

def message_from_string(s: AnyStr, *args, **kwargs): ...
def message_from_bytes(s: str, *args, **kwargs): ...
def message_from_file(fp: IO[AnyStr], *args, **kwargs): ...
def message_from_binary_file(fp: IO[str], *args, **kwargs): ...



from overrides import overrides

from .array_end_node import ArrayEndNode
from .array_start_widget import ArrayStartWidget
from .array_unexpanded_widget import ArrayUnexpandedWidget
from ..json_composite_end_node import JSONCompositeEndNode
from ..json_composite_node import JSONCompositeNode


class ArrayNode(JSONCompositeNode):
    """
    implementation of JSON `array` type node
    """

    def __init__(self, key: str, value: list, node_factory,
                 parent=None, display_key=True):
        super().__init__(key, value, node_factory, parent, display_key)
        self._children = {}
        self._size = len(value)

    @overrides
    def collapse_all(self):
        for index, child in self._children.items():
            if isinstance(child, (JSONCompositeNode, JSONCompositeEndNode)):
                child.collapse_all()
        if self.is_expanded():
            self.toggle_expanded()

    @overrides
    def has_children(self):
        return self._size != 0

    @overrides
    def get_first_child(self):
        if not self.has_children():
            return None
        return self._get_child_node(0)

    @overrides
    def get_last_child(self):
        if not self.has_children():
            return None
        return self._get_child_node(self._size - 1)

    @overrides
    def prev_child(self, key):
        index = int(key)
        if index == 0:
            return None
        return self._get_child_node(index - 1)

    @overrides
    def next_child(self, key):
        index = int(key)
        if index == self._size - 1:
            return None
        return self._get_child_node(index + 1)

    def _get_child_node(self, index):
        if not self.has_children():
            return None

        if index < 0 or index >= self._size:
            raise TypeError(
                f"index ${index} is out of bound ${self._size}in {type(self)}"
                f"#get_child_node."
            )
        elif index not in self._children:
            self._children[index] = self._load_child_node(index)

        return self._children[index]

    def _load_child_node(self, index):
        value = self.get_value()[index]
        return self._node_factory.create_node(
            str(index), value, parent=self, display_key=False
        )

    # ======================================================================= #
    # ui                                                                      #
    # ======================================================================= #

    @overrides
    def load_start_widget(self):
        return ArrayStartWidget(self, self.is_display_key())

    @overrides
    def load_unexpanded_widget(self):
        return ArrayUnexpandedWidget(self, self.is_display_key())

    @overrides
    def load_end_node(self):
        return ArrayEndNode(self)



from __future__ import unicode_literals

__all__ = (
    'InputMode',
    'CharacterFind',
    'ViState',
)


class InputMode(object):
    INSERT = 'vi-insert'
    INSERT_MULTIPLE = 'vi-insert-multiple'
    NAVIGATION = 'vi-navigation'
    REPLACE = 'vi-replace'


class CharacterFind(object):
    def __init__(self, character, backwards=False):
        self.character = character
        self.backwards = backwards


class ViState(object):
    """
    Mutable class to hold the state of the Vi navigation.
    """
    def __init__(self):
        #: None or CharacterFind instance. (This is used to repeat the last
        #: search in Vi mode, by pressing the 'n' or 'N' in navigation mode.)
        self.last_character_find = None

        # When an operator is given and we are waiting for text object,
        # -- e.g. in the case of 'dw', after the 'd' --, an operator callback
        # is set here.
        self.operator_func = None
        self.operator_arg = None

        #: Named registers. Maps register name (e.g. 'a') to
        #: :class:`ClipboardData` instances.
        self.named_registers = {}

        #: The Vi mode we're currently in to.
        self.input_mode = InputMode.INSERT

        #: Waiting for digraph.
        self.waiting_for_digraph = False
        self.digraph_symbol1 = None  # (None or a symbol.)

        #: When true, make ~ act as an operator.
        self.tilde_operator = False

    def reset(self, mode=InputMode.INSERT):
        """
        Reset state, go back to the given mode. INSERT by default.
        """
        # Go back to insert mode.
        self.input_mode = mode

        self.waiting_for_digraph = False
        self.operator_func = None
        self.operator_arg = None



from copy import deepcopy
import re
from termcolor import colored
from types import (
    BuiltinMethodType,
    BuiltinFunctionType,
    FunctionType,
    MethodType,
    ModuleType,
)
import sys


def peep(obj, builtins=False, privates=False, docstrings=False,
         truncate_len=250):
    if not isinstance(obj, ModuleType):
        obj = deepcopy(obj)
    obj_dir = dir(obj)
    output = OutputStrWrapper(truncate_len)
    if builtins is False:
        _is_magic = lambda x: (x.startswith("__") and x.endswith("__"))
        obj_dir = [x for x in obj_dir if not _is_magic(x)]

    if privates is False:
        _is_private = lambda x: (x.startswith("_") and not x.endswith("_"))
        obj_dir = [x for x in obj_dir if not _is_private(x)]

    # check if debugging and define output function
    gettrace = getattr(sys, "gettrace", None)
    if gettrace():
        debug = True
    else:
        debug = False

    print(colored(getattr(obj, "__name__", ""), "red"))
    doc = getattr(obj, "__doc__", "")
    if callable(obj):
        try:
            print(obj())
        except Exception as e:
            if _positional_exception(e):
                msg = "(requires positional arguments)"
                msg_color = "grey"
            else:
                msg = f"RAISES EXCEPTION : {e}"
                msg_color = "red"
            output.add(
                colored(f"{obj}(): ", "magenta"),
                colored(msg, msg_color),
                colored(doc, "green"),
            )
            print(colored(msg, "grey"))
        print(colored(obj.__doc__, "green"))

    for item in obj_dir:
        attr = getattr(obj, item)
        doc = ""
        try:
            if docstrings and attr.__doc__ is not None:
                doc = "\n" + attr.__doc__
            eval_str = str(attr)
            eval_str = "" if eval_str == "None" else eval_str
            if callable(attr) is False:
                eval_str = _fix_if_multiline(eval_str)
                output.add(
                    colored(f"{item}: ", "cyan"),
                    eval_str,
                    colored(doc, "green")
                )
        except (Exception, BaseException) as e:
            output.add(
                colored(f"{item}: ", "cyan"),
                colored(f"RAISES EXCEPTION : {e}", "red"),
                colored(doc, "green"),
            )
            continue
        if callable(attr):
            try:
                msg = str(attr())
                msg = _fix_if_multiline(msg)
                output.add(
                    colored(f"{item}(): ", "magenta"),
                    msg,
                    colored(doc, "green")
                )
            except (Exception, BaseException) as e:
                if _positional_exception(e):
                    msg = "(requires positional arguments)"
                    msg_color = "grey"
                else:
                    msg = f"RAISES EXCEPTION : {e}"
                    msg_color = "red"
                output.add(
                    colored(f"{item}(): ", "magenta"),
                    colored(msg, msg_color),
                    colored(doc, "green"),
                )
                continue
    return output


class OutputStrWrapper:
    def __init__(self, max_len):
        self._max_len = max_len
        self._output_str = ''

    def add(self, *args):
        str_ = _shorten(*args, max_len=self._max_len)
        self._output_str += str_ + '\n'

    def __repr__(self):
        return str(self._output_str)

    def __str__(self):
        return repr(self)


class Peeper:
    def __init__(
        self, obj, builtins=False, privates=False, docstrings=False, truncate_len=250
    ):
        pass


def _shorten(*args, max_len):
    str_ = "".join(args)
    if max_len:
        if len(str_) > max_len:
            str_ = str_[:max_len] + " ..."
    return str_


def _fix_if_multiline(msg):
    if "\n" in msg:
        return "\n" + msg
    else:
        return msg


def _positional_exception(e):
    e = str(e).lower()
    msgs = ('required argument', 'positional argument', 'missing argument',
            'requires argument', 'must be given', 'takes exactly')
    re_msgs = (r'exactly . argument', r'at least .* argument', )
    if any([msg in e for msg in msgs]):
        return True
    if any([re.compile(msg).search(e) for msg in re_msgs]):
        return True
    else:
        return False


if __name__ == '__main__':
    print(peep('asdf'))
    import numpy as np
    arr = np.array([1, 2, 3])
    print(peep(arr))



# Stubs for sunau (Python 2 and 3)

import sys
from typing import Any, NamedTuple, NoReturn, Optional, Text, IO, Union, Tuple

_File = Union[Text, IO[bytes]]

class Error(Exception): ...

AUDIO_FILE_MAGIC: int
AUDIO_FILE_ENCODING_MULAW_8: int
AUDIO_FILE_ENCODING_LINEAR_8: int
AUDIO_FILE_ENCODING_LINEAR_16: int
AUDIO_FILE_ENCODING_LINEAR_24: int
AUDIO_FILE_ENCODING_LINEAR_32: int
AUDIO_FILE_ENCODING_FLOAT: int
AUDIO_FILE_ENCODING_DOUBLE: int
AUDIO_FILE_ENCODING_ADPCM_G721: int
AUDIO_FILE_ENCODING_ADPCM_G722: int
AUDIO_FILE_ENCODING_ADPCM_G723_3: int
AUDIO_FILE_ENCODING_ADPCM_G723_5: int
AUDIO_FILE_ENCODING_ALAW_8: int
AUDIO_UNKNOWN_SIZE: int

if sys.version_info < (3, 0):
    _sunau_params = Tuple[int, int, int, int, str, str]
else:
    _sunau_params = NamedTuple('_sunau_params', [
        ('nchannels', int),
        ('sampwidth', int),
        ('framerate', int),
        ('nframes', int),
        ('comptype', str),
        ('compname', str),
    ])

class Au_read:
    def __init__(self, f: _File) -> None: ...
    if sys.version_info >= (3, 3):
        def __enter__(self) -> Au_read: ...
        def __exit__(self, *args: Any) -> None: ...
    def getfp(self) -> Optional[IO[bytes]]: ...
    def rewind(self) -> None: ...
    def close(self) -> None: ...
    def tell(self) -> int: ...
    def getnchannels(self) -> int: ...
    def getnframes(self) -> int: ...
    def getsampwidth(self) -> int: ...
    def getframerate(self) -> int: ...
    def getcomptype(self) -> str: ...
    def getcompname(self) -> str: ...
    def getparams(self) -> _sunau_params: ...
    def getmarkers(self) -> None: ...
    def getmark(self, id: Any) -> NoReturn: ...
    def setpos(self, pos: int) -> None: ...
    def readframes(self, nframes: int) -> Optional[bytes]: ...

class Au_write:
    def __init__(self, f: _File) -> None: ...
    if sys.version_info >= (3, 3):
        def __enter__(self) -> Au_write: ...
        def __exit__(self, *args: Any) -> None: ...
    def setnchannels(self, nchannels: int) -> None: ...
    def getnchannels(self) -> int: ...
    def setsampwidth(self, sampwidth: int) -> None: ...
    def getsampwidth(self) -> int: ...
    def setframerate(self, framerate: float) -> None: ...
    def getframerate(self) -> int: ...
    def setnframes(self, nframes: int) -> None: ...
    def getnframes(self) -> int: ...
    def setcomptype(self, comptype: str, compname: str) -> None: ...
    def getcomptype(self) -> str: ...
    def getcompname(self) -> str: ...
    def setparams(self, params: _sunau_params) -> None: ...
    def getparams(self) -> _sunau_params: ...
    def setmark(self, id: Any, pos: Any, name: Any) -> NoReturn: ...
    def getmark(self, id: Any) -> NoReturn: ...
    def getmarkers(self) -> None: ...
    def tell(self) -> int: ...
    # should be any bytes-like object after 3.4, but we don't have a type for that
    def writeframesraw(self, data: bytes) -> None: ...
    def writeframes(self, data: bytes) -> None: ...
    def close(self) -> None: ...

# Returns a Au_read if mode is rb and Au_write if mode is wb
def open(f: _File, mode: Optional[str] = ...) -> Any: ...
openfp = open



import abc
import sys
from typing import (
    Dict, Type, TypeVar, Optional, Union, Any, Generic, Mapping, ItemsView, KeysView, ValuesView
)

_T = TypeVar('_T')
_U = TypeVar('_U')

# Internal mypy fallback type for all typed dicts (does not exist at runtime)
class _TypedDict(Mapping[str, object], metaclass=abc.ABCMeta):
    def copy(self: _T) -> _T: ...
    # Using NoReturn so that only calls using mypy plugin hook that specialize the signature
    # can go through.
    def setdefault(self, k: NoReturn, default: object) -> object: ...
    # Mypy plugin hook for 'pop' expects that 'default' has a type variable type.
    def pop(self, k: NoReturn, default: _T = ...) -> object: ...
    def update(self: _T, __m: _T) -> None: ...
    if sys.version_info < (3, 0):
        def has_key(self, k: str) -> bool: ...
        def viewitems(self) -> ItemsView[str, object]: ...
        def viewkeys(self) -> KeysView[str]: ...
        def viewvalues(self) -> ValuesView[object]: ...
    def __delitem__(self, k: NoReturn) -> None: ...

def TypedDict(typename: str, fields: Dict[str, Type[_T]], total: bool = ...) -> Type[dict]: ...

def Arg(type: _T = ..., name: Optional[str] = ...) -> _T: ...
def DefaultArg(type: _T = ..., name: Optional[str] = ...) -> _T: ...
def NamedArg(type: _T = ..., name: Optional[str] = ...) -> _T: ...
def DefaultNamedArg(type: _T = ..., name: Optional[str] = ...) -> _T: ...
def VarArg(type: _T = ...) -> _T: ...
def KwArg(type: _T = ...) -> _T: ...

# Return type that indicates a function does not return.
# This type is equivalent to the None type, but the no-op Union is necessary to
# distinguish the None type from the None value.
NoReturn = Union[None]  # Deprecated: Use typing.NoReturn instead.

# This is intended as a class decorator, but mypy rejects abstract classes
# when a Type[_T] is expected, so we can't give it the type we want
def trait(cls: Any) -> Any: ...

class FlexibleAlias(Generic[_T, _U]): ...



from enum import Enum

import urwid
from overrides import overrides

from ...json_lib import JSONListBox
from ...json_lib import JSONListWalker
from ...json_lib import NodeFactory
from ...json_lib import DEFAULT_NODE_IMPLS


class JSONBrowserKeys(Enum):
    # keys for json lib
    CURSOR_UP = "up"
    CURSOR_DOWN = "down"
    COLLAPSE_ALL = "c"
    EXPAND_ALL = "e"
    TOGGLE_EXPANSION = "enter"

    # keys for switching window
    OPEN_QUERY_BAR = "."


class JSONBrowser(urwid.WidgetWrap):
    """
    Window to display JSON contents.
    """

    def __init__(self, mediator, keymapper, data=""):
        self._keymapper = keymapper
        self._mediator = mediator
        self._node_factory = NodeFactory(DEFAULT_NODE_IMPLS)
        self._top_node = self._node_factory.create_root_node(data)

        super().__init__(self._load_widget())
        self._mediator.register("refresh_view", self.set_top_node)

    def set_top_node(self, data):
        self._top_node = self._node_factory.create_node(
            "", data, display_key=False
        )
        self._refresh()

    def _load_widget(self):
        listbox = JSONListBox(JSONListWalker(self._top_node))
        return urwid.AttrMap(listbox, "body")

    def _refresh(self):
        self._w = self._load_widget()

    @overrides
    def keypress(self, size, key):
        key = self._keymapper.key(key)
        key = super().keypress(size, key)

        if key == JSONBrowserKeys.OPEN_QUERY_BAR.value:
            self._mediator.notify("open_query_bar", "json_browser")
            return

        return key



from copy import copy, deepcopy
import rp
import time
import abc
import subprocess

from pynput import mouse, keyboard
from pynput.mouse import Controller as MouseController, Button
from pynput.keyboard import Controller as KeyboardController, Key

##############################################################################
# A C T I O N   C L A S S E S
##############################################################################

mouse_controller = MouseController()
keyboard_controller = KeyboardController()


class Reprable:
    def __repr__(self):
        attributes = [
            f"{attr}={str(value)}"
            for attr, value in self.__dict__.items()
            if not "__" in attr and value != []
        ]
        return f'{type(self).__name__}({", ".join(attributes)})'


class Action(abc.ABC, Reprable):
    """
    Base class for all actions. Each action is callable.
    """

    @abc.abstractmethod
    def __call__(self):
        """
        Perform the action, e.g. press a key, move the mouse, etc.
        """
        pass


class DelayAction(Action):
    """
    Sleeps for `seconds` when called.
    """

    def __init__(self, seconds: float):
        self.seconds = seconds

    def __call__(self):
        if self.seconds > 0:
            time.sleep(self.seconds)


# ------------------- Mouse Actions -------------------
class MouseAction(Action):
    """
    Base for mouse actions: provides x, y coordinates.
    """

    def __init__(self, x, y):
        self.x = x
        self.y = y


class MouseMoveAction(MouseAction):
    """
    Moves the mouse pointer to (x, y).
    """

    def __init__(self, x, y):
        super().__init__(x, y)

    def __call__(self):
        mouse_controller.position = (self.x, self.y)


class MousePressAction(MouseAction):
    """
    Presses a mouse button at (x, y).
    """

    def __init__(self, x, y, button=Button.left):
        super().__init__(x, y)
        self.button = button

    def __call__(self):
        mouse_controller.position = (self.x, self.y)
        mouse_controller.press(self.button)


class MouseReleaseAction(MouseAction):
    """
    Releases a mouse button at (x, y).
    """

    def __init__(self, x, y, button=Button.left):
        super().__init__(x, y)
        self.button = button

    def __call__(self):
        mouse_controller.position = (self.x, self.y)
        mouse_controller.release(self.button)


class MouseScrollAction(MouseAction):
    """
    Scrolls the mouse wheel at (x, y) by (dx, dy).
    """

    def __init__(self, x, y, dx, dy):
        super().__init__(x, y)
        self.dx = dx
        self.dy = dy

    def __call__(self):
        mouse_controller.position = (self.x, self.y)
        mouse_controller.scroll(self.dx, self.dy)


# ------------------- Keyboard Actions -------------------
class KeyboardAction(Action):
    """
    Base for keyboard actions. Stores a Key or a character.
    """

    def __init__(self, key):
        self.key = key


class KeyboardPressAction(KeyboardAction):
    """
    Presses a key (character or special Key).
    """

    def __init__(self, key):
        super().__init__(key)

    def __call__(self):
        keyboard_controller.press(self.key)


class KeyboardReleaseAction(KeyboardAction):
    """
    Releases a key (character or special Key).
    """

    def __init__(self, key):
        super().__init__(key)

    def __call__(self):
        keyboard_controller.release(self.key)


##############################################################################
# T R I G G E R S
##############################################################################


class Trigger(abc.ABC, Reprable):
    """
    Each trigger is a callable that takes an Action object.
    It returns True if triggered, otherwise False.
    Triggers can manage internal state as needed.
    """

    NUM_ACTIONS = 0

    @abc.abstractmethod
    def __call__(self, action: Action) -> bool:
        pass


class NoStartTrigger(Trigger):
    """
    Always returns True immediately -> recording starts right away.
    """

    def __call__(self, action: Action) -> bool:
        return True


class NoEndTrigger(Trigger):
    """
    Never returns True -> must stop via Ctrl+C or other means.
    """

    def __call__(self, action: Action) -> bool:
        return False


class KeyDownTrigger(Trigger):
    NUM_ACTIONS = 1

    def __init__(self, target_key):
        self.target_key = target_key

    def __call__(self, action: Action) -> bool:
        return (
            isinstance(action, KeyboardReleaseAction) and action.key == self.target_key
        )


class KeyUpTrigger(Trigger):
    NUM_ACTIONS = 1

    def __init__(self, target_key):
        self.target_key = target_key

    def __call__(self, action: Action) -> bool:
        return isinstance(action, KeyboardPressAction) and action.key == self.target_key


class TwoKeyDownTrigger(Trigger):
    NUM_ACTIONS = 2

    def __init__(self, key1, key2):
        self.key1 = key1
        self.key2 = key2
        self.key1_pressed = False
        self.key2_pressed = False
        self.triggered = False

    def __call__(self, action: Action) -> bool:
        if self.triggered:
            return False

        if isinstance(action, KeyboardPressAction):
            if action.key in self.key1:
                self.key1_pressed = True
            elif action.key in self.key2:
                self.key2_pressed = True
            else:
                self.key1_pressed = False
                self.key2_pressed = False

            if self.key1_pressed and self.key2_pressed:
                self.triggered = True
                return True

        elif isinstance(action, KeyboardReleaseAction):
            self.key1_pressed = False
            self.key2_pressed = False

        return False


class TwoKeyUpTrigger(Trigger):
    NUM_ACTIONS = 2

    def __init__(self, key1, key2):
        self.key1 = key1
        self.key2 = key2
        self.key1_released = False
        self.key2_released = False
        self.down_trigger = TwoKeyDownTrigger(self.key1, self.key2)
        self.triggered = False
        self.ready = False

    def __call__(self, action: Action) -> bool:
        if self.triggered:
            return False

        self.ready |= self.down_trigger(action)

        if self.ready and isinstance(action, KeyboardReleaseAction):
            self.key1_released |= action.key in self.key1
            self.key2_released |= action.key in self.key2

            if self.ready and self.key1_released and self.key2_released:
                self.triggered = True
                return True

        #FOR DEBUGGING
        # print(
        #     line_join(
        #         [
        #             fansi(self, "white green"),
        #             fansi(action, "yellow white"),
        #             fansi(self.down_trigger, "orange"),
        #         ]
        #     )
        # )

        return False


class CmdEscUpTrigger(TwoKeyUpTrigger):
    def __init__(self):
        super().__init__([Key.cmd, Key.cmd_r], [Key.esc])


class CmdEscDownTrigger(TwoKeyDownTrigger):
    def __init__(self):
        super().__init__([Key.cmd, Key.cmd_r], [Key.esc])


class CmdShiftUpTrigger(TwoKeyUpTrigger):
    def __init__(self):
        super().__init__([Key.cmd, Key.cmd_r], [Key.shift])


class CmdShiftDownTrigger(TwoKeyDownTrigger):
    def __init__(self):
        super().__init__([Key.cmd, Key.cmd_r], [Key.shift])


class CmdCmdUpTrigger(TwoKeyUpTrigger):
    # This is buggy because pynput is buggy with respect to two cmd keys! Don't use
    def __init__(self):
        super().__init__([Key.cmd], [Key.cmd_r])


class CmdCmdDownTrigger(TwoKeyDownTrigger):
    def __init__(self):
        super().__init__([Key.cmd], [Key.cmd_r])


##############################################################################
# R E C O R D I N G  W I T H  S E P A R A T E  D E L A Y S
##############################################################################


def record_actions(start_trigger=None, end_trigger=None):
    """
    Capture a list of Action objects:
      - DelayAction(seconds)
      - MouseMoveAction, MousePressAction, etc.
      - KeyboardPressAction, KeyboardReleaseAction
    The action that triggers start or end is NOT recorded.
    """
    if start_trigger is None:
        start_trigger = NoStartTrigger()
    if end_trigger is None:
        end_trigger = NoEndTrigger()

    actions = []
    start_recording = False
    stop_recording = False
    last_time = None

    def create_action(event_type, data):
        """
        Convert a (event_type, data) from pynput into one of our Action objects.
        """
        if event_type == "mouse_move":
            x, y = data
            return MouseMoveAction(x, y)

        elif event_type == "mouse_press":
            btn, (x, y) = data
            button = Button.left if str(btn) == "Button.left" else Button.right
            return MousePressAction(x, y, button)

        elif event_type == "mouse_release":
            btn, (x, y) = data
            button = Button.left if str(btn) == "Button.left" else Button.right
            return MouseReleaseAction(x, y, button)

        elif event_type == "mouse_scroll":
            x, y, dx, dy = data
            return MouseScrollAction(x, y, dx, dy)

        elif event_type == "keyboard_press":
            return KeyboardPressAction(data)

        elif event_type == "keyboard_release":
            return KeyboardReleaseAction(data)

        return None  # Unknown event

    def handle_event(event_type, data):
        nonlocal stop_recording

        if stop_recording:
            return

        nonlocal start_recording, last_time

        # Convert to an Action object
        action = create_action(event_type, data)
        if action is None:
            return  # skip unknown events

        # Check triggers

        if not start_recording:
            rp.fansi_print(action, "italic dark gray")
            if start_trigger(action):
                # Just triggered the start
                start_recording = True
                last_time = time.time()
            return  # Don't add actions before starting!
        else:
            if end_trigger(action):
                # Just triggered the end
                print("BLOOBAH")
                for _ in range(end_trigger.NUM_ACTIONS):
                    while actions and isinstance(actions[-1], DelayAction):
                        actions.pop()
                    if actions:
                        actions.pop()
                        deleted_action = actions.pop()
                        # This is ugly logic...actions should instead be FILTERED through triggers.
                        # For this reason, I'll abstract this whole codebase away into an rp function, as I might improve this logic in the future to handle new types of triggers.
                        rp.fansi_print(
                            f"[{len(actions)}]: ACTIONS -= {action}", "orange red blue green"
                        )
                stop_recording = True
                return

        # If we are here, we are actively recording this action
        now = time.time()
        if last_time is None:
            last_time = now
        delta = now - last_time
        last_time = now

        # Add a DelayAction if delta > 0
        if delta > 0:
            actions.append(DelayAction(delta))

        # Add the actual action
        actions.append(action)
        rp.fansi_print(f"[{len(actions)}]: ACTIONS += {action}", "yellow blue green")

    # ---------------------------------------------------------------------
    #  Listener callbacks
    # ---------------------------------------------------------------------
    def on_move(x, y):
        handle_event("mouse_move", (x, y))

    def on_click(x, y, button, pressed):
        if pressed:
            handle_event("mouse_press", (button, (x, y)))
        else:
            handle_event("mouse_release", (button, (x, y)))

    def on_scroll(x, y, dx, dy):
        handle_event("mouse_scroll", (x, y, dx, dy))

    def on_press(key):
        handle_event("keyboard_press", key)

    def on_release(key):
        handle_event("keyboard_release", key)

    # Start listeners
    mouse_listener = mouse.Listener(
        on_move=on_move, on_click=on_click, on_scroll=on_scroll
    )
    keyboard_listener = keyboard.Listener(on_press=on_press, on_release=on_release)
    mouse_listener.start()
    keyboard_listener.start()

    print("Recording...Press cmd+shift anywhere or Ctrl+C in the terminal to stop.")
    try:
        while not stop_recording:
            time.sleep(0.000001)
    except KeyboardInterrupt:
        print("Stopped by KeyboardInterrupt.")
    finally:
        mouse_listener.stop()
        keyboard_listener.stop()

    print(f"Recorded {len(actions)} actions (including DelayActions).")
    return actions

##############################################################################
# T M U X
##############################################################################

# This section was generated with the help of Claude-Code, May 14 2025

# Mapping from pynput Key to tmux key name
PYNPUT_TO_TMUX_KEYS = {
    Key.alt: "M",         # Only for tracking active modifiers
    Key.alt_r: "M",       # Only for tracking active modifiers
    Key.backspace: "BSpace",
    Key.ctrl: "C",        # Only for tracking active modifiers
    Key.ctrl_r: "C",      # Only for tracking active modifiers
    Key.delete: "DC",
    Key.down: "Down",
    Key.end: "End",
    Key.enter: "Enter",
    Key.esc: "Escape",
    Key.f1: "F1",
    Key.f2: "F2",
    Key.f3: "F3",
    Key.f4: "F4",
    Key.f5: "F5",
    Key.f6: "F6",
    Key.f7: "F7",
    Key.f8: "F8",
    Key.f9: "F9",
    Key.f10: "F10",
    Key.f11: "F11",
    Key.f12: "F12",
    Key.f13: "F13",
    Key.f14: "F14",
    Key.f15: "F15",
    Key.f16: "F16",
    Key.f17: "F17",
    Key.f18: "F18",
    Key.f19: "F19",
    Key.f20: "F20",
    Key.home: "Home",
    Key.left: "Left",
    Key.page_down: "PageDown", 
    Key.page_up: "PageUp",
    Key.right: "Right",
    Key.space: "Space",
    Key.tab: "Tab",
    Key.up: "Up",
}

# Special characters that need escaping in tmux
TMUX_SPECIAL_CHARS = {
    ';': '\\;',
    '#': '\\#',
    ',': '\\,',
}

# macOS special characters to their original Alt/Option keystroke combinations
# These are the characters that appear when using Alt/Option + key combinations
MACOS_SPECIAL_CHAR_TO_KEYSTROKE = {
    # Option + letter combinations
    'Ω': ('M', 'z'),        # Option + z
    'π': ('M', 'p'),        # Option + p  
    'µ': ('M', 'm'),        # Option + m
    'ß': ('M', 's'),        # Option + s
    'ç': ('M', 'c'),        # Option + c
    '√': ('M', 'v'),        # Option + v
    '≈': ('M', 'x'),        # Option + x
    '∂': ('M', 'd'),        # Option + d
    '∆': ('M', 'j'),        # Option + j
    '∫': ('M', 'b'),        # Option + b
    '∑': ('M', 'w'),        # Option + w
    '∞': ('M', '5'),        # Option + 5
    '≤': ('M', ','),        # Option + comma
    '≥': ('M', '.'),        # Option + period
    '÷': ('M', '/'),        # Option + /
    '≠': ('M', '='),        # Option + =
    '±': ('M', '+'),        # Option + +
    '¬': ('M', 'l'),        # Option + l
    '∧': ('M', 'k'),        # Option + k
    '∨': ('M', 'v'),        # Option + v (logical or)
    '⊕': ('M', 'q'),        # Option + q
    '⊗': ('M', 'r'),        # Option + r
    '⊙': ('M', 'o'),        # Option + o
    '⊥': ('M', 'i'),        # Option + i
    '∪': ('M', 'u'),        # Option + u
    '∩': ('M', 'n'),        # Option + n
    '∅': ('M', 'h'),        # Option + h
    '∈': ('M', 'e'),        # Option + e
    '∉': ('M', 'f'),        # Option + f
    '∋': ('M', 'g'),        # Option + g
    '∌': ('M', 't'),        # Option + t
    '∠': ('M', 'a'),        # Option + a
    '£': ('M', '3'),        # Option + 3
    '¢': ('M', '4'),        # Option + 4
    '¥': ('M', 'y'),        # Option + y
    '¡': ('M', '1'),        # Option + 1
    '«': ('M', '\\'),       # Option + \
    '"': ('M', '['),        # Option + [
    "'": ('M', ']'),        # Option + ]
    '–': ('M', '-'),        # Option + -
    '…': ('M', ';'),        # Option + ;
    '†': ('M', 't'),        # Option + t
    '®': ('M', 'r'),        # Option + r
    '´': ('M', 'e'),        # Option + e (acute accent dead key)
    'œ': ('M', 'q'),        # Option + q
    'å': ('M', 'a'),        # Option + a
    'ƒ': ('M', 'f'),        # Option + f
    '©': ('M', 'g'),        # Option + g
    'ø': ('M', 'o'),        # Option + o
    'æ': ('M', "'"),        # Option + '
    '`': ('M', '`'),        # Option + `
    'ˆ': ('M', 'i'),        # Option + i (circumflex accent dead key)
    '™': ('M', '2'),        # Option + 2
    '§': ('M', '6'),        # Option + 6
    '¶': ('M', '7'),        # Option + 7
    '•': ('M', '8'),        # Option + 8
    'ª': ('M', '9'),        # Option + 9
    'º': ('M', '0'),        # Option + 0
    
    # Option + Shift combinations
    'Ç': ('M', 'S-c'),      # Option + Shift + c
    '◊': ('M', 'S-v'),      # Option + Shift + v
    '˛': ('M', 'S-x'),      # Option + Shift + x
    '¸': ('M', 'S-z'),      # Option + Shift + z
    '˝': ('M', 'S-w'),      # Option + Shift + w
    '˚': ('M', 'S-k'),      # Option + Shift + k
    '¯': ('M', 'S-h'),      # Option + Shift + h
    '˘': ('M', 'S-g'),      # Option + Shift + g
    '˙': ('M', 'S-h'),      # Option + Shift + h
    '¨': ('M', 'S-u'),      # Option + Shift + u
    '˜': ('M', 'S-n'),      # Option + Shift + n
    '¦': ('M', 'S-\\'),     # Option + Shift + \
    '€': ('M', 'S-2'),      # Option + Shift + 2
    '‰': ('M', 'S-r'),      # Option + Shift + r
    '°': ('M', 'S-8'),      # Option + Shift + 8
    '¿': ('M', 'S-/'),      # Option + Shift + /
    '»': ('M', 'S-\\'),     # Option + Shift + \
    '"': ('M', 'S-['),      # Option + Shift + [
    "'": ('M', 'S-]'),      # Option + Shift + ]
    '—': ('M', 'S--'),      # Option + Shift + -
    '‚': ('M', 'S-;'),      # Option + Shift + ;
    '„': ('M', 'S-w'),      # Option + Shift + w
    '‹': ('M', 'S-3'),      # Option + Shift + 3
    '›': ('M', 'S-4'),      # Option + Shift + 4
    'ﬁ': ('M', 'S-5'),      # Option + Shift + 5
    'ﬂ': ('M', 'S-6'),      # Option + Shift + 6
    '‡': ('M', 'S-7'),      # Option + Shift + 7
    '·': ('M', 'S-9'),      # Option + Shift + 9
    '⁄': ('M', 'S-1'),      # Option + Shift + 1
    'Œ': ('M', 'S-q'),      # Option + Shift + q
    'Å': ('M', 'S-a'),      # Option + Shift + a
    'Í': ('M', 'S-s'),      # Option + Shift + s
    'Î': ('M', 'S-d'),      # Option + Shift + d
    'Ï': ('M', 'S-f'),      # Option + Shift + f
    'Ó': ('M', 'S-h'),      # Option + Shift + h (duplicate with ¯)
    'Ô': ('M', 'S-j'),      # Option + Shift + j
    'Ò': ('M', 'S-k'),      # Option + Shift + k (duplicate with ˚)
    'Ú': ('M', 'S-l'),      # Option + Shift + l
    'Æ': ('M', 'S-\\'),     # Option + Shift + " (quote key)
    'Á': ('M', 'S-e'),      # Option + Shift + e
    'ˇ': ('M', 'S-t'),      # Option + Shift + t
    '´': ('M', 'S-e'),      # Option + Shift + e (duplicate)
    'Â': ('M', 'S-i'),      # Option + Shift + i
    'ı': ('M', 'S-b'),      # Option + Shift + b
    'Ø': ('M', 'S-o'),      # Option + Shift + o
    '∏': ('M', 'S-p'),      # Option + Shift + p
    '\u2019': ('M', 'S-]'),      # Option + Shift + ] (right single quote)
    '\u2018': ('M', 'S-['),      # Option + Shift + [ (left single quote)
    '\u201C': ('M', 'S-['),      # Option + Shift + [ (left double quote)
    '\u201D': ('M', 'S-]'),      # Option + Shift + ] (right double quote)
}


def _convert_macos_special_char(char, cls):
    """
    Helper function to convert macOS special characters to TmuxSendkeyAction.
    
    Args:
        char: The character to check and convert
        cls: The TmuxSendkeyAction class
        
    Returns:
        TmuxSendkeyAction or None: The converted action if char is a special character, None otherwise
    """
    if char in MACOS_SPECIAL_CHAR_TO_KEYSTROKE:
        modifier, keystroke = MACOS_SPECIAL_CHAR_TO_KEYSTROKE[char]
        # For special characters, we ignore the active_modifiers and use the mapped ones
        if keystroke.startswith('S-'):
            # Option + Shift combination
            return cls(keystroke[2:], [modifier, 'S'])
        else:
            # Option only combination
            return cls(keystroke, [modifier])
    return None


class TmuxSendkeyAction(Action):
    """
    Represents a tmux send-keys command.
    Can be initialized with a tmux key string or converted from a pynput key.
    """
    
    def __init__(self, key_str, modifiers=None):
        """
        Initialize with a tmux key string (e.g., "Enter", "Space", "C-c").
        
        Args:
            key_str (str): The key string in tmux format
            modifiers (list): Optional list of modifier keys ['C', 'M', 'S']
                              C = Control, M = Alt/Meta, S = Shift
        """
        self.key_str = key_str
        self.modifiers = modifiers or []
        
    def __call__(self):
        """
        Execute the tmux send-keys command to send this key to tmux.
        """
        
        # Tmux format for key combinations varies based on key type
        if self.modifiers:
            # For special keys with modifiers (like function keys, arrows, etc.)
            if self.key_str in ["Enter", "Space", "BSpace", "DC", "Down", "Up", "Left", "Right", 
                               "Home", "End", "PageUp", "PageDown", "Tab", "Escape"] or \
               (self.key_str.startswith("F") and self.key_str[1:].isdigit()):
                # Format as: prefix key_name (e.g., "C-M-Up")
                mod_prefix = '-'.join(self.modifiers)
                key_arg = f"{mod_prefix}-{self.key_str}"
                args = ["tmux", "send-keys", key_arg]
            else:
                # Regular character with modifiers
                mod_prefix = '-'.join(self.modifiers)
                key_arg = f"{mod_prefix}-{self.key_str}"
                args = ["tmux", "send-keys", key_arg]
        else:
            # For special keys without modifiers, use the name format
            if self.key_str in ["Enter", "Space", "BSpace", "DC", "Down", "Up", "Left", "Right", 
                               "Home", "End", "PageUp", "PageDown", "Tab", "Escape"] or \
               (self.key_str.startswith("F") and self.key_str[1:].isdigit()):
                args = ["tmux", "send-keys", self.key_str]
            else:
                # Regular character
                args = ["tmux", "send-keys", self.key_str]
                
        cmd_str = " ".join(args)
        
        try:
            # Actually execute the tmux command
            subprocess.run(args, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Error executing tmux command: {e}")
        except FileNotFoundError:
            print("Error: tmux command not found. Is tmux installed and in your PATH?")
    
    @classmethod
    def from_pynput(cls, key, active_modifiers=None, strict=True):
        """
        Convert a pynput key to a TmuxSendkeyAction.
        
        Args:
            key: The pynput key to convert (Key enum, KeyCode, or char)
            active_modifiers (dict): Dictionary with current active modifiers
                                    e.g. {'ctrl': True, 'alt': False, 'shift': False}
            strict (bool): If True, raise an exception if the key can't be converted;
                          otherwise, return None
        
        Returns:
            TmuxSendkeyAction or None: The converted tmux action
            
        Raises:
            ValueError: If the key can't be converted and strict is True
        """
        modifiers = []
        if active_modifiers:
            if active_modifiers.get('ctrl'):
                modifiers.append('C')
            if active_modifiers.get('alt'):
                modifiers.append('M')
            if active_modifiers.get('shift'):
                modifiers.append('S')
        
        # Special keys (like Key.enter, Key.space, etc.)
        if isinstance(key, Key):
            # Don't convert modifier keys themselves when they're pressed
            if key in [Key.ctrl, Key.ctrl_r, Key.alt, Key.alt_r, 
                      Key.shift, Key.shift_r]:
                return None
                
            tmux_key = PYNPUT_TO_TMUX_KEYS.get(key)
            if tmux_key is None:
                if strict:
                    raise ValueError(f"Cannot convert pynput key {key} to tmux format")
                return None
            return cls(tmux_key, modifiers)
        
        # Regular characters: could be str or KeyCode object
        if hasattr(key, 'char') and key.char is not None:
            # KeyCode with character representation
            char = key.char
            
            # Check for macOS special characters first
            if result := _convert_macos_special_char(char, cls):
                return result
                    
            if char in TMUX_SPECIAL_CHARS:
                return cls(TMUX_SPECIAL_CHARS[char], modifiers)
            return cls(char, modifiers)
            
        # String literals
        elif isinstance(key, str):
            # Check for macOS special characters first
            if result := _convert_macos_special_char(key, cls):
                return result
                    
            # Escape special characters
            if key in TMUX_SPECIAL_CHARS:
                return cls(TMUX_SPECIAL_CHARS[key], modifiers)
            # Regular character
            return cls(key, modifiers)
        
        # Try to get string representation as fallback
        try:
            key_str = str(key)
            # Check if it's a quoted character like 'a'
            if key_str.startswith("'") and key_str.endswith("'") and len(key_str) == 3:
                char = key_str[1]
                
                # Check for macOS special characters first
                if result := _convert_macos_special_char(char, cls):
                    return result
                        
                if char in TMUX_SPECIAL_CHARS:
                    return cls(TMUX_SPECIAL_CHARS[char], modifiers)
                return cls(char, modifiers)
        except:
            pass
        
        # Unknown key type
        if strict:
            raise ValueError(f"Unknown key type: {type(key)} - {key}")
        return None

def pynput_casette_to_tmux(casette, *, strict):
    """
    Convert the PynputCasette to a list of TmuxSendkeyAction objects.
    
    This conversion:
    1. Keeps all DelayAction objects
    2. Converts KeyboardPressAction to TmuxSendkeyAction, including modifiers
    3. Removes all KeyboardReleaseAction objects
    4. Removes all mouse-related actions
    5. Tracks active modifiers (ctrl, alt, shift) for proper key combinations
    
    Args:
        strict (bool): If True, raise an exception if any key can't be converted;
                      if False, skip keys that can't be converted
                      
    Returns:
        PynputCasette: A new casette containing only DelayAction and TmuxSendkeyAction objects
        
    Raises:
        ValueError: If strict=True and a key can't be converted
    """
    tmux_actions = []
    
    # Track which modifier keys are currently pressed
    active_modifiers = {
        'ctrl': False,
        'alt': False,
        'shift': False,
    }
    
    for action in casette:
        if isinstance(action, (DelayAction, TmuxSendkeyAction)):
            # Keep all delay actions and tmux actions
            tmux_actions.append(copy(action))

        elif isinstance(action, KeyboardPressAction):
            # Update modifiers state
            if action.key in [Key.ctrl, Key.ctrl_r]:
                active_modifiers['ctrl'] = True
            elif action.key in [Key.alt, Key.alt_r]:
                active_modifiers['alt'] = True
            elif action.key in [Key.shift, Key.shift_r]:
                active_modifiers['shift'] = True
            else:
                # Try to convert keyboard press actions to tmux actions with current modifiers
                try:
                    tmux_action = TmuxSendkeyAction.from_pynput(
                        action.key, 
                        active_modifiers=active_modifiers,
                        strict=strict
                    )
                    if tmux_action is not None:
                        tmux_actions.append(tmux_action)
                except ValueError as e:
                    if strict:
                        raise ValueError(f"Failed to convert action {action}: {e}")
                    
        elif isinstance(action, KeyboardReleaseAction):
            # Update modifiers state on release
            if action.key in [Key.ctrl, Key.ctrl_r]:
                active_modifiers['ctrl'] = False
            elif action.key in [Key.alt, Key.alt_r]:
                active_modifiers['alt'] = False
            elif action.key in [Key.shift, Key.shift_r]:
                active_modifiers['shift'] = False
                
        # Skip all other actions (mouse actions)
        
    # Merge consecutive delay actions for cleaner output
    return PynputCasette(tmux_actions).merge_delays()

##############################################################################
# P L A Y B A C K
##############################################################################

def playback_actions(actions):
    """
    Replays each Action in sequence by simply calling it.
    Press escape to interrupt playback!
    """

    #A safekey
    abort=False
    abort_key=Key.esc
    def on_press(key):
        nonlocal abort
        if key==abort_key:
            abort=True
            safekey_listener.stop()
    safekey_listener = keyboard.Listener(on_press=on_press)
    safekey_listener.start()

    rp.fansi_print(f"Playing back {len(actions)} actions...press {abort_key} to abort!",'white white bold')
    for i, action in enumerate(actions, start=1):
        if abort:
            rp.fansi_print('Aborting playback!','red orange','bold')
            raise KeyboardInterrupt
        action()
        style = None
        if isinstance(action, DelayAction): style = 'dark gray italic'
        if isinstance(action, MouseMoveAction): style = 'dark light gray'
        elif isinstance(action, MouseAction): style = 'gray green yellow bold'
        if isinstance(action, KeyboardReleaseAction): style = 'white blue red blue magenta bold'
        if isinstance(action, KeyboardPressAction): style = 'white blue green blue cyan bold'
        if isinstance(action, TmuxSendkeyAction): style = 'light orange gray orange bold'
        action_str = rp.fansi(action, style)
        num_str = rp.fansi(f"[{i}/{len(actions)}]", "yellow gray")
        print(f"{num_str} {action_str}")

    rp.fansi_print("Playback finished.",'white white bold')


##############################################################################
# D E L A Y   U T I L I T I E S
##############################################################################


def lstrip_delays(actions):
    """
    If the first action is a DelayAction, set its .seconds = 0.
    """
    while actions and isinstance(actions[0], DelayAction):
        actions = actions[1:]
    return actions


def rstrip_delays(actions):
    """
    If the last action is a DelayAction, set its .seconds = 0.
    """
    while actions and isinstance(actions[-1], DelayAction):
        actions = actions[:-1]
    return actions


def strip_delays(actions):
    """
    Set the first and last DelayAction (if any) to 0.0
    """
    actions = lstrip_delays(actions)
    actions = rstrip_delays(actions)
    return actions


##############################################################################
# Default Recorder
##############################################################################


class PynputCasette(list):
    def __init__(self, actions=None):
        super().__init__([] if actions is None else actions)
        self.start_trigger = CmdShiftUpTrigger()
        self.end_trigger = CmdShiftDownTrigger()

    def record(self):
        return PynputCasette(
            self + record_actions(self.start_trigger, self.end_trigger)
        )

    def strip(self):
        return PynputCasette(strip_delays(copy(self)))

    def lstrip(self):
        return PynputCasette(lstrip_delays(copy(self)))

    def rstrip(self):
        return PynputCasette(rstrip_delays(copy(self)))

    def speedup(self, factor):
        self = deepcopy(self)
        for x in self:
            if isinstance(x, DelayAction):
                x.seconds /= factor
        return self

    def slowdown(self, factor):
        return self.speedup(1 / factor)

    def remove_mousemoves(self):
        return PynputCasette(x for x in self if not isinstance(x, MouseMoveAction)).merge_delays()

    def merge_delays(self):
        out = []
        for x in self:
            x = copy(x)
            if isinstance(x, DelayAction) and out and isinstance(out[-1], DelayAction):
                out[-1].seconds+=x.seconds
            else:
                out.append(x)
        return PynputCasette(out)

    def set_all_delays(self,func_or_float):
        """ Takes a float -> float func OR a float"""
        if callable(func_or_float):
            func = func_or_float
        else:
            func_or_float = float(func_or_float)
            func = lambda x: func_or_float

        out = []
        for x in self:
            x=copy(x)
            if isinstance(x, DelayAction):
                x.seconds=func(x.seconds)
            out.append(x)
        return PynputCasette(out)

    def cap_all_delays(self, seconds):
        return self.set_all_delays(lambda x: min(x, seconds))

    def round_all_delays(self, ndigits=2):
        """ Rounds all delays to ndigits decimals; good for compressing via rp.object_to_base64 """
        return self.set_all_delays(lambda x: round(x, ndigits))

    def to_tmux(self, *, strict=True):
        return pynput_casette_to_tmux(self, strict=strict)

    @property
    def duration(self):
        return sum(x.seconds for x in self if isinstance(x, DelayAction))

    def play(self, loop=False):
        for _ in range(1 + 999**999 * loop):
            playback_actions(self)

    def repr(self):
        return f"{type(self).__name__}({repr(self.actions)}"




import random
import copy

from ... import colors
from ... import console
from ... import nbinput

from ... import canvas
from ... import sprite
from ... import shapes

BOTTOM, DOWN, LEFT, TOP, UP, RIGHT = 0, 0, 1, 2, 2, 3

Canvas = canvas.Canvas
Sprite = sprite.Sprite
NonBlockingInput = nbinput.NonBlockingInput
BlockingInput = nbinput.BlockingInput



from typing import Any

class SMTPException(Exception): ...
class SMTPServerDisconnected(SMTPException): ...

class SMTPResponseException(SMTPException):
    smtp_code: Any
    smtp_error: Any
    args: Any
    def __init__(self, code, msg) -> None: ...

class SMTPSenderRefused(SMTPResponseException):
    smtp_code: Any
    smtp_error: Any
    sender: Any
    args: Any
    def __init__(self, code, msg, sender) -> None: ...

class SMTPRecipientsRefused(SMTPException):
    recipients: Any
    args: Any
    def __init__(self, recipients) -> None: ...

class SMTPDataError(SMTPResponseException): ...
class SMTPConnectError(SMTPResponseException): ...
class SMTPHeloError(SMTPResponseException): ...
class SMTPAuthenticationError(SMTPResponseException): ...

def quoteaddr(addr): ...
def quotedata(data): ...

class SSLFakeFile:
    sslobj: Any
    def __init__(self, sslobj) -> None: ...
    def readline(self, size=...): ...
    def close(self): ...

class SMTP:
    debuglevel: Any
    file: Any
    helo_resp: Any
    ehlo_msg: Any
    ehlo_resp: Any
    does_esmtp: Any
    default_port: Any
    timeout: Any
    esmtp_features: Any
    local_hostname: Any
    def __init__(self, host: str = ..., port: int = ..., local_hostname=..., timeout=...) -> None: ...
    def set_debuglevel(self, debuglevel): ...
    sock: Any
    def connect(self, host=..., port=...): ...
    def send(self, str): ...
    def putcmd(self, cmd, args=...): ...
    def getreply(self): ...
    def docmd(self, cmd, args=...): ...
    def helo(self, name=...): ...
    def ehlo(self, name=...): ...
    def has_extn(self, opt): ...
    def help(self, args=...): ...
    def rset(self): ...
    def noop(self): ...
    def mail(self, sender, options=...): ...
    def rcpt(self, recip, options=...): ...
    def data(self, msg): ...
    def verify(self, address): ...
    vrfy: Any
    def expn(self, address): ...
    def ehlo_or_helo_if_needed(self): ...
    def login(self, user, password): ...
    def starttls(self, keyfile=..., certfile=...): ...
    def sendmail(self, from_addr, to_addrs, msg, mail_options=..., rcpt_options=...): ...
    def close(self): ...
    def quit(self): ...

class SMTP_SSL(SMTP):
    default_port: Any
    keyfile: Any
    certfile: Any
    def __init__(self, host=..., port=..., local_hostname=..., keyfile=..., certfile=..., timeout=...) -> None: ...

class LMTP(SMTP):
    ehlo_msg: Any
    def __init__(self, host=..., port=..., local_hostname=...) -> None: ...
    sock: Any
    def connect(self, host=..., port=...): ...



# Stubs for site

from typing import List, Iterable, Optional
import sys

PREFIXES: List[str]
ENABLE_USER_SITE: Optional[bool]
USER_SITE: Optional[str]
USER_BASE: Optional[str]

if sys.version_info < (3,):
    def main() -> None: ...
def addsitedir(sitedir: str,
               known_paths: Optional[Iterable[str]] = ...) -> None: ...
def getsitepackages(prefixes: Optional[Iterable[str]] = ...) -> List[str]: ...
def getuserbase() -> str: ...
def getusersitepackages() -> str: ...



# Stubs for ssl

from typing import (
    Any, Dict, Callable, List, NamedTuple, Optional, Set, Tuple, Union,
)
import socket
import sys

_PCTRTT = Tuple[Tuple[str, str], ...]
_PCTRTTT = Tuple[_PCTRTT, ...]
_PeerCertRetDictType = Dict[str, Union[str, _PCTRTTT, _PCTRTT]]
_PeerCertRetType = Union[_PeerCertRetDictType, bytes, None]
_EnumRetType = List[Tuple[bytes, str, Union[Set[str], bool]]]
_PasswordType = Union[Callable[[], Union[str, bytes]], str, bytes]

if sys.version_info >= (3, 5):
    _SC1ArgT = Union[SSLSocket, SSLObject]
else:
    _SC1ArgT = SSLSocket
_SrvnmeCbType = Callable[[_SC1ArgT, Optional[str], SSLSocket], Optional[int]]

class SSLError(OSError):
    library: str
    reason: str
class SSLZeroReturnError(SSLError): ...
class SSLWantReadError(SSLError): ...
class SSLWantWriteError(SSLError): ...
class SSLSyscallError(SSLError): ...
class SSLEOFError(SSLError): ...

if sys.version_info >= (3, 7):
    class SSLCertVerificationError(SSLError, ValueError):
        verify_code: int
        verify_message: str

    CertificateError = SSLCertVerificationError
else:
    class CertificateError(ValueError): ...


def wrap_socket(sock: socket.socket, keyfile: Optional[str] = ...,
                certfile: Optional[str] = ..., server_side: bool = ...,
                cert_reqs: int = ..., ssl_version: int = ...,
                ca_certs: Optional[str] = ...,
                do_handshake_on_connect: bool = ...,
                suppress_ragged_eofs: bool = ...,
                ciphers: Optional[str] = ...) -> SSLSocket: ...


if sys.version_info < (3,) or sys.version_info >= (3, 4):
    def create_default_context(purpose: Any = ..., *,
                               cafile: Optional[str] = ...,
                               capath: Optional[str] = ...,
                               cadata: Optional[str] = ...) -> SSLContext: ...

if sys.version_info >= (3, 4):
    def _create_unverified_context(protocol: int = ..., *,
                                   cert_reqs: int = ...,
                                   check_hostname: bool = ...,
                                   purpose: Any = ...,
                                   certfile: Optional[str] = ...,
                                   keyfile: Optional[str] = ...,
                                   cafile: Optional[str] = ...,
                                   capath: Optional[str] = ...,
                                   cadata: Optional[str] = ...) -> SSLContext: ...
    _create_default_https_context: Callable[..., SSLContext]

if sys.version_info >= (3, 3):
    def RAND_bytes(num: int) -> bytes: ...
    def RAND_pseudo_bytes(num: int) -> Tuple[bytes, bool]: ...
def RAND_status() -> bool: ...
def RAND_egd(path: str) -> None: ...
def RAND_add(bytes: bytes, entropy: float) -> None: ...


def match_hostname(cert: _PeerCertRetType, hostname: str) -> None: ...
def cert_time_to_seconds(cert_time: str) -> int: ...
def get_server_certificate(addr: Tuple[str, int], ssl_version: int = ...,
                           ca_certs: Optional[str] = ...) -> str: ...
def DER_cert_to_PEM_cert(der_cert_bytes: bytes) -> str: ...
def PEM_cert_to_DER_cert(pem_cert_string: str) -> bytes: ...
if sys.version_info < (3,) or sys.version_info >= (3, 4):
    DefaultVerifyPaths = NamedTuple('DefaultVerifyPaths',
                                    [('cafile', str), ('capath', str),
                                     ('openssl_cafile_env', str),
                                     ('openssl_cafile', str),
                                     ('openssl_capath_env', str),
                                     ('openssl_capath', str)])
    def get_default_verify_paths() -> DefaultVerifyPaths: ...

if sys.platform == 'win32':
    if sys.version_info < (3,) or sys.version_info >= (3, 4):
        def enum_certificates(store_name: str) -> _EnumRetType: ...
        def enum_crls(store_name: str) -> _EnumRetType: ...


CERT_NONE: int
CERT_OPTIONAL: int
CERT_REQUIRED: int

if sys.version_info < (3,) or sys.version_info >= (3, 4):
    VERIFY_DEFAULT: int
    VERIFY_CRL_CHECK_LEAF: int
    VERIFY_CRL_CHECK_CHAIN: int
    VERIFY_X509_STRICT: int
    VERIFY_X509_TRUSTED_FIRST: int

PROTOCOL_SSLv23: int
PROTOCOL_SSLv2: int
PROTOCOL_SSLv3: int
PROTOCOL_TLSv1: int
if sys.version_info < (3,) or sys.version_info >= (3, 4):
    PROTOCOL_TLSv1_1: int
    PROTOCOL_TLSv1_2: int
if sys.version_info >= (3, 5):
    PROTOCOL_TLS: int
if sys.version_info >= (3, 6):
    PROTOCOL_TLS_CLIENT: int
    PROTOCOL_TLS_SERVER: int

OP_ALL: int
OP_NO_SSLv2: int
OP_NO_SSLv3: int
OP_NO_TLSv1: int
if sys.version_info < (3,) or sys.version_info >= (3, 4):
    OP_NO_TLSv1_1: int
    OP_NO_TLSv1_2: int
OP_CIPHER_SERVER_PREFERENCE: int
OP_SINGLE_DH_USE: int
OP_SINGLE_ECDH_USE: int
OP_NO_COMPRESSION: int
if sys.version_info >= (3, 6):
    OP_NO_TICKET: int

if sys.version_info < (3,) or sys.version_info >= (3, 5):
    HAS_ALPN: int
HAS_ECDH: bool
HAS_SNI: bool
HAS_NPN: bool
CHANNEL_BINDING_TYPES: List[str]

OPENSSL_VERSION: str
OPENSSL_VERSION_INFO: Tuple[int, int, int, int, int]
OPENSSL_VERSION_NUMBER: int

if sys.version_info < (3,) or sys.version_info >= (3, 4):
    ALERT_DESCRIPTION_HANDSHAKE_FAILURE: int
    ALERT_DESCRIPTION_INTERNAL_ERROR: int
    ALERT_DESCRIPTION_ACCESS_DENIED: int
    ALERT_DESCRIPTION_BAD_CERTIFICATE: int
    ALERT_DESCRIPTION_BAD_CERTIFICATE_HASH_VALUE: int
    ALERT_DESCRIPTION_BAD_CERTIFICATE_STATUS_RESPONSE: int
    ALERT_DESCRIPTION_BAD_RECORD_MAC: int
    ALERT_DESCRIPTION_CERTIFICATE_EXPIRED: int
    ALERT_DESCRIPTION_CERTIFICATE_REVOKED: int
    ALERT_DESCRIPTION_CERTIFICATE_UNKNOWN: int
    ALERT_DESCRIPTION_CERTIFICATE_UNOBTAINABLE: int
    ALERT_DESCRIPTION_CLOSE_NOTIFY: int
    ALERT_DESCRIPTION_DECODE_ERROR: int
    ALERT_DESCRIPTION_DECOMPRESSION_FAILURE: int
    ALERT_DESCRIPTION_DECRYPT_ERROR: int
    ALERT_DESCRIPTION_ILLEGAL_PARAMETER: int
    ALERT_DESCRIPTION_INSUFFICIENT_SECURITY: int
    ALERT_DESCRIPTION_NO_RENEGOTIATION: int
    ALERT_DESCRIPTION_PROTOCOL_VERSION: int
    ALERT_DESCRIPTION_RECORD_OVERFLOW: int
    ALERT_DESCRIPTION_UNEXPECTED_MESSAGE: int
    ALERT_DESCRIPTION_UNKNOWN_CA: int
    ALERT_DESCRIPTION_UNKNOWN_PSK_IDENTITY: int
    ALERT_DESCRIPTION_UNRECOGNIZED_NAME: int
    ALERT_DESCRIPTION_UNSUPPORTED_CERTIFICATE: int
    ALERT_DESCRIPTION_UNSUPPORTED_EXTENSION: int
    ALERT_DESCRIPTION_USER_CANCELLED: int

if sys.version_info < (3,) or sys.version_info >= (3, 4):
    _PurposeType = NamedTuple('_PurposeType', [('nid', int), ('shortname', str), ('longname', str), ('oid', str)])
    class Purpose:
        SERVER_AUTH: _PurposeType
        CLIENT_AUTH: _PurposeType


class SSLSocket(socket.socket):
    context: SSLContext
    server_side: bool
    server_hostname: Optional[str]
    if sys.version_info >= (3, 6):
        session: Optional[SSLSession]
        session_reused: Optional[bool]

    def read(self, len: int = ...,
             buffer: Optional[bytearray] = ...) -> bytes: ...
    def write(self, buf: bytes) -> int: ...
    def do_handshake(self) -> None: ...
    def getpeercert(self, binary_form: bool = ...) -> _PeerCertRetType: ...
    def cipher(self) -> Tuple[str, int, int]: ...
    if sys.version_info >= (3, 5):
        def shared_cipher(self) -> Optional[List[Tuple[str, int, int]]]: ...
    def compression(self) -> Optional[str]: ...
    def get_channel_binding(self, cb_type: str = ...) -> Optional[bytes]: ...
    if sys.version_info < (3,) or sys.version_info >= (3, 5):
        def selected_alpn_protocol(self) -> Optional[str]: ...
    def selected_npn_protocol(self) -> Optional[str]: ...
    def unwrap(self) -> socket.socket: ...
    if sys.version_info < (3,) or sys.version_info >= (3, 5):
        def version(self) -> Optional[str]: ...
    def pending(self) -> int: ...


class SSLContext:
    if sys.version_info < (3,) or sys.version_info >= (3, 4):
        check_hostname: bool
    options: int
    @property
    def protocol(self) -> int: ...
    if sys.version_info < (3,) or sys.version_info >= (3, 4):
        verify_flags: int
    verify_mode: int
    if sys.version_info >= (3, 5):
        def __init__(self, protocol: int = ...) -> None: ...
    else:
        def __init__(self, protocol: int) -> None: ...
    if sys.version_info < (3,) or sys.version_info >= (3, 4):
        def cert_store_stats(self) -> Dict[str, int]: ...
    def load_cert_chain(self, certfile: str, keyfile: Optional[str] = ...,
                        password: _PasswordType = ...) -> None: ...
    if sys.version_info < (3,) or sys.version_info >= (3, 4):
        def load_default_certs(self, purpose: _PurposeType = ...) -> None: ...
        def load_verify_locations(self, cafile: Optional[str] = ...,
                                  capath: Optional[str] = ...,
                                  cadata: Union[str, bytes, None] = ...) -> None: ...
        def get_ca_certs(self,
                         binary_form: bool = ...) -> Union[List[_PeerCertRetDictType], List[bytes]]: ...
    else:
        def load_verify_locations(self,
                                  cafile: Optional[str] = ...,
                                  capath: Optional[str] = ...) -> None: ...
    def set_default_verify_paths(self) -> None: ...
    def set_ciphers(self, ciphers: str) -> None: ...
    if sys.version_info < (3,) or sys.version_info >= (3, 5):
        def set_alpn_protocols(self, protocols: List[str]) -> None: ...
    def set_npn_protocols(self, protocols: List[str]) -> None: ...
    def set_servername_callback(self,
                                server_name_callback: Optional[_SrvnmeCbType]) -> None: ...
    def load_dh_params(self, dhfile: str) -> None: ...
    def set_ecdh_curve(self, curve_name: str) -> None: ...
    def wrap_socket(self, sock: socket.socket, server_side: bool = ...,
                    do_handshake_on_connect: bool = ...,
                    suppress_ragged_eofs: bool = ...,
                    server_hostname: Optional[str] = ...) -> SSLSocket: ...
    if sys.version_info >= (3, 5):
        def wrap_bio(self, incoming: MemoryBIO, outgoing: MemoryBIO,
                     server_side: bool = ...,
                     server_hostname: Optional[str] = ...) -> SSLObject: ...
    def session_stats(self) -> Dict[str, int]: ...


if sys.version_info >= (3, 5):
    class SSLObject:
        context: SSLContext
        server_side: bool
        server_hostname: Optional[str]
        if sys.version_info >= (3, 6):
            session: Optional[SSLSession]
            session_reused: bool
        def read(self, len: int = ...,
                 buffer: Optional[bytearray] = ...) -> bytes: ...
        def write(self, buf: bytes) -> int: ...
        def getpeercert(self, binary_form: bool = ...) -> _PeerCertRetType: ...
        def selected_npn_protocol(self) -> Optional[str]: ...
        def cipher(self) -> Tuple[str, int, int]: ...
        def shared_cipher(self) -> Optional[List[Tuple[str, int, int]]]: ...
        def compression(self) -> Optional[str]: ...
        def pending(self) -> int: ...
        def do_handshake(self) -> None: ...
        def unwrap(self) -> None: ...
        def get_channel_binding(self, cb_type: str = ...) -> Optional[bytes]: ...

    class MemoryBIO:
        pending: int
        eof: bool
        def read(self, n: int = ...) -> bytes: ...
        def write(self, buf: bytes) -> int: ...
        def write_eof(self) -> None: ...

if sys.version_info >= (3, 6):
    class SSLSession:
        id: bytes
        time: int
        timeout: int
        ticket_lifetime_hint: int
        has_ticket: bool


# TODO below documented in cpython but not in docs.python.org
# taken from python 3.4
SSL_ERROR_EOF: int
SSL_ERROR_INVALID_ERROR_CODE: int
SSL_ERROR_SSL: int
SSL_ERROR_SYSCALL: int
SSL_ERROR_WANT_CONNECT: int
SSL_ERROR_WANT_READ: int
SSL_ERROR_WANT_WRITE: int
SSL_ERROR_WANT_X509_LOOKUP: int
SSL_ERROR_ZERO_RETURN: int

def get_protocol_name(protocol_code: int) -> str: ...

AF_INET: int
PEM_FOOTER: str
PEM_HEADER: str
SOCK_STREAM: int
SOL_SOCKET: int
SO_TYPE: int



import time
import os
import sys
import hashlib
import gc
import shutil
import platform
import errno
import logging

try:
    import cPickle as pickle
except:
    import pickle

from rp.libs.parso._compatibility import FileNotFoundError

LOG = logging.getLogger(__name__)


_PICKLE_VERSION = 32
"""
Version number (integer) for file system cache.

Increment this number when there are any incompatible changes in
the parser tree classes.  For example, the following changes
are regarded as incompatible.

- A class name is changed.
- A class is moved to another module.
- A __slot__ of a class is changed.
"""

_VERSION_TAG = '%s-%s%s-%s' % (
    platform.python_implementation(),
    sys.version_info[0],
    sys.version_info[1],
    _PICKLE_VERSION
)
"""
Short name for distinguish Python implementations and versions.

It's like `sys.implementation.cache_tag` but for Python < 3.3
we generate something similar.  See:
http://docs.python.org/3/library/sys.html#sys.implementation
"""


def _get_default_cache_path():
    if platform.system().lower() == 'windows':
        dir_ = os.path.join(os.getenv('LOCALAPPDATA') or '~', 'Parso', 'Parso')
    elif platform.system().lower() == 'darwin':
        dir_ = os.path.join('~', 'Library', 'Caches', 'Parso')
    else:
        dir_ = os.path.join(os.getenv('XDG_CACHE_HOME') or '~/.cache', 'parso')
    return os.path.expanduser(dir_)


_default_cache_path = _get_default_cache_path()
"""
The path where the cache is stored.

On Linux, this defaults to ``~/.cache/parso/``, on OS X to
``~/Library/Caches/Parso/`` and on Windows to ``%LOCALAPPDATA%\\Parso\\Parso\\``.
On Linux, if environment variable ``$XDG_CACHE_HOME`` is set,
``$XDG_CACHE_HOME/parso`` is used instead of the default one.
"""

parser_cache = {}


class _NodeCacheItem(object):
    def __init__(self, node, lines, change_time=None):
        self.node = node
        self.lines = lines
        if change_time is None:
            change_time = time.time()
        self.change_time = change_time


def load_module(hashed_grammar, file_io, cache_path=None):
    """
    Returns a module or None, if it fails.
    """
    p_time = file_io.get_last_modified()
    if p_time is None:
        return None

    try:
        module_cache_item = parser_cache[hashed_grammar][file_io.path]
        if p_time <= module_cache_item.change_time:
            return module_cache_item.node
    except KeyError:
        return _load_from_file_system(
            hashed_grammar,
            file_io.path,
            p_time,
            cache_path=cache_path
        )


def _load_from_file_system(hashed_grammar, path, p_time, cache_path=None):
    cache_path = _get_hashed_path(hashed_grammar, path, cache_path=cache_path)
    try:
        try:
            if p_time > os.path.getmtime(cache_path):
                # Cache is outdated
                return None
        except OSError as e:
            if e.errno == errno.ENOENT:
                # In Python 2 instead of an IOError here we get an OSError.
                raise FileNotFoundError
            else:
                raise

        with open(cache_path, 'rb') as f:
            gc.disable()
            try:
                module_cache_item = pickle.load(f)
            finally:
                gc.enable()
    except FileNotFoundError:
        return None
    else:
        parser_cache.setdefault(hashed_grammar, {})[path] = module_cache_item
        LOG.debug('pickle loaded: %s', path)
        return module_cache_item.node


def save_module(hashed_grammar, file_io, module, lines, pickling=True, cache_path=None):
    path = file_io.path
    try:
        p_time = None if path is None else file_io.get_last_modified()
    except OSError:
        p_time = None
        pickling = False

    item = _NodeCacheItem(module, lines, p_time)
    parser_cache.setdefault(hashed_grammar, {})[path] = item
    if pickling and path is not None:
        _save_to_file_system(hashed_grammar, path, item, cache_path=cache_path)


def _save_to_file_system(hashed_grammar, path, item, cache_path=None):
    with open(_get_hashed_path(hashed_grammar, path, cache_path=cache_path), 'wb') as f:
        pickle.dump(item, f, pickle.HIGHEST_PROTOCOL)


def clear_cache(cache_path=None):
    if cache_path is None:
        cache_path = _default_cache_path
    shutil.rmtree(cache_path)
    parser_cache.clear()


def _get_hashed_path(hashed_grammar, path, cache_path=None):
    directory = _get_cache_directory_path(cache_path=cache_path)

    file_hash = hashlib.sha256(path.encode("utf-8")).hexdigest()
    return os.path.join(directory, '%s-%s.pkl' % (hashed_grammar, file_hash))


def _get_cache_directory_path(cache_path=None):
    if cache_path is None:
        cache_path = _default_cache_path
    directory = os.path.join(cache_path, _VERSION_TAG)
    if not os.path.exists(directory):
        os.makedirs(directory)
    return directory



from typing import Any, Optional
from boto.compat import StandardError

class BotoClientError(StandardError):
    reason: Any
    def __init__(self, reason, *args) -> None: ...

class SDBPersistenceError(StandardError): ...
class StoragePermissionsError(BotoClientError): ...
class S3PermissionsError(StoragePermissionsError): ...
class GSPermissionsError(StoragePermissionsError): ...

class BotoServerError(StandardError):
    status: Any
    reason: Any
    body: Any
    request_id: Any
    error_code: Any
    message: str
    box_usage: Any
    def __init__(self, status, reason, body: Optional[Any] = ..., *args) -> None: ...
    def __getattr__(self, name): ...
    def __setattr__(self, name, value): ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...

class ConsoleOutput:
    parent: Any
    instance_id: Any
    timestamp: Any
    comment: Any
    output: Any
    def __init__(self, parent: Optional[Any] = ...) -> None: ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...

class StorageCreateError(BotoServerError):
    bucket: Any
    def __init__(self, status, reason, body: Optional[Any] = ...) -> None: ...
    def endElement(self, name, value, connection): ...

class S3CreateError(StorageCreateError): ...
class GSCreateError(StorageCreateError): ...
class StorageCopyError(BotoServerError): ...
class S3CopyError(StorageCopyError): ...
class GSCopyError(StorageCopyError): ...

class SQSError(BotoServerError):
    detail: Any
    type: Any
    def __init__(self, status, reason, body: Optional[Any] = ...) -> None: ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...

class SQSDecodeError(BotoClientError):
    message: Any
    def __init__(self, reason, message) -> None: ...

class StorageResponseError(BotoServerError):
    resource: Any
    def __init__(self, status, reason, body: Optional[Any] = ...) -> None: ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...

class S3ResponseError(StorageResponseError): ...
class GSResponseError(StorageResponseError): ...

class EC2ResponseError(BotoServerError):
    errors: Any
    def __init__(self, status, reason, body: Optional[Any] = ...) -> None: ...
    def startElement(self, name, attrs, connection): ...
    request_id: Any
    def endElement(self, name, value, connection): ...

class JSONResponseError(BotoServerError):
    status: Any
    reason: Any
    body: Any
    error_message: Any
    error_code: Any
    def __init__(self, status, reason, body: Optional[Any] = ..., *args) -> None: ...

class DynamoDBResponseError(JSONResponseError): ...
class SWFResponseError(JSONResponseError): ...
class EmrResponseError(BotoServerError): ...

class _EC2Error:
    connection: Any
    error_code: Any
    error_message: Any
    def __init__(self, connection: Optional[Any] = ...) -> None: ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...

class SDBResponseError(BotoServerError): ...
class AWSConnectionError(BotoClientError): ...
class StorageDataError(BotoClientError): ...
class S3DataError(StorageDataError): ...
class GSDataError(StorageDataError): ...

class InvalidUriError(Exception):
    message: Any
    def __init__(self, message) -> None: ...

class InvalidAclError(Exception):
    message: Any
    def __init__(self, message) -> None: ...

class InvalidCorsError(Exception):
    message: Any
    def __init__(self, message) -> None: ...

class NoAuthHandlerFound(Exception): ...

class InvalidLifecycleConfigError(Exception):
    message: Any
    def __init__(self, message) -> None: ...

class ResumableTransferDisposition:
    START_OVER: str
    WAIT_BEFORE_RETRY: str
    ABORT_CUR_PROCESS: str
    ABORT: str

class ResumableUploadException(Exception):
    message: Any
    disposition: Any
    def __init__(self, message, disposition) -> None: ...

class ResumableDownloadException(Exception):
    message: Any
    disposition: Any
    def __init__(self, message, disposition) -> None: ...

class TooManyRecordsException(Exception):
    message: Any
    def __init__(self, message) -> None: ...

class PleaseRetryException(Exception):
    message: Any
    response: Any
    def __init__(self, message, response: Optional[Any] = ...) -> None: ...

class InvalidInstanceMetadataError(Exception):
    MSG: str
    def __init__(self, msg) -> None: ...



"""
Lexer interface and implementation.
Used for syntax highlighting.
"""
from __future__ import unicode_literals
from abc import ABCMeta, abstractmethod
from six import with_metaclass
from six.moves import range

from rp.prompt_toolkit.token import Token
from rp.prompt_toolkit.filters import to_cli_filter
from .utils import split_lines

import re
import six

# <CLAUDE CODE START: Lazy imports for multi-language highlighting - imports deferred until needed>
# Lexer imports moved inside LazyLexer classes to avoid slow startup
# <CLAUDE CODE END: Lazy imports for multi-language highlighting>

__all__ = (
    'Lexer',
    'SimpleLexer',
    'PygmentsLexer',
    'LazyPygmentsLexer',
    'SyntaxSync',
    'SyncFromStart',
    'RegexSync',
)


class Lexer(with_metaclass(ABCMeta, object)):
    """
    Base class for all lexers.
    """
    @abstractmethod
    def lex_document(self, cli, document):
        """
        Takes a :class:`~prompt_toolkit.document.Document` and returns a
        callable that takes a line number and returns the tokens for that line.
        """


class SimpleLexer(Lexer):
    """
    Lexer that doesn't do any tokenizing and returns the whole input as one token.

    :param token: The `Token` for this lexer.
    """
    # `default_token` parameter is deprecated!
    def __init__(self, token=Token, default_token=None):
        self.token = token

        if default_token is not None:
            self.token = default_token

    def lex_document(self, cli, document):
        lines = document.lines

        def get_line(lineno):
            " Return the tokens for the given line. "
            try:
                return [(self.token, lines[lineno])]
            except IndexError:
                return []
        return get_line


class SyntaxSync(with_metaclass(ABCMeta, object)):
    """
    Syntax synchroniser. This is a tool that finds a start position for the
    lexer. This is especially important when editing big documents; we don't
    want to start the highlighting by running the lexer from the beginning of
    the file. That is very slow when editing.
    """
    @abstractmethod
    def get_sync_start_position(self, document, lineno):
        """
        Return the position from where we can start lexing as a (row, column)
        tuple.

        :param document: `Document` instance that contains all the lines.
        :param lineno: The line that we want to highlight. (We need to return
            this line, or an earlier position.)
        """

class SyncFromStart(SyntaxSync):
    """
    Always start the syntax highlighting from the beginning.
    """
    def get_sync_start_position(self, document, lineno):
        return 0, 0


class RegexSync(SyntaxSync):
    """
    Synchronize by starting at a line that matches the given regex pattern.
    """
    # Never go more than this amount of lines backwards for synchronisation.
    # That would be too CPU intensive.
    MAX_BACKWARDS = 500

    # Start lexing at the start, if we are in the first 'n' lines and no
    # synchronisation position was found.
    FROM_START_IF_NO_SYNC_POS_FOUND = 100

    def __init__(self, pattern):
        assert isinstance(pattern, six.text_type)
        self._compiled_pattern = re.compile(pattern)

    def get_sync_start_position(self, document, lineno):
        " Scan backwards, and find a possible position to start. "
        pattern = self._compiled_pattern
        lines = document.lines

        # Scan upwards, until we find a point where we can start the syntax
        # synchronisation.
        for i in range(lineno, max(-1, lineno - self.MAX_BACKWARDS), -1):
            match = pattern.match(lines[i])
            if match:
                return i, match.start()

        # No synchronisation point found. If we aren't that far from the
        # beginning, start at the very beginning, otherwise, just try to start
        # at the current line.
        if lineno < self.FROM_START_IF_NO_SYNC_POS_FOUND:
            return 0, 0
        else:
            return lineno, 0

    @classmethod
    def from_pygments_lexer_cls(cls, lexer_cls):
        """
        Create a :class:`.RegexSync` instance for this Pygments lexer class.
        """
        patterns = {
            # For Python, start highlighting at any class/def block.
            'Python':   r'^\s*(class|def)\s+',
            'Python 3': r'^\s*(class|def)\s+',

            # For HTML, start at any open/close tag definition.
            'HTML': r'<[/a-zA-Z]',

            # For javascript, start at a function.
            'JavaScript': r'\bfunction\b'

            # TODO: Add definitions for other languages.
            #       By default, we start at every possible line.
        }
        p = patterns.get(lexer_cls.name, '^')
        return cls(p)


def longest_common_prefix(a,b):
    #Written by Ryan Burgert, 2020. Written for efficiency's sake.
    #Works for strings, lists and tuples (and possibly other datatypes, but not numpy arrays)
    #This implementation is two orders of magnitude faster than anything I could find on the web/stack overflow/etc, especially for strings
    #It has complexity O(len(output of this function)), and a very good time constant (because it doesn't directly iterate through every element in a python loop)
    #On my computer, this function was able to compare two strings of length 1,000,000 in 0.00454 second. Here's the test I used: string='a'*10**7;tic();longest_common_prefix(string,string);ptoc() [[[tic() starts a timer, ptoc() prints out the elapsed time]]]
    #
    #EXAMPLES:
    #   longest_common_prefix('abcderty','abcdefoaisjd')                --> abcde
    #   longest_common_prefix('abcderty','abcsdefoa')                   --> abc
    #   longest_common_prefix('abcderty','asbcsdefoa')                  --> a
    #   longest_common_prefix('abcderty','aasbcsdefoa')                 --> a
    #   longest_common_prefix('aaaabdcderty','aasbcsdefoa')             --> aa
    #   longest_common_prefix(list('aaaabdcderty'),list('aasbcsdefoa')) --> ['a', 'a']
    
    len_a=len(a)
    len_b=len(b)
    out_max=min(len_a,len_b)
    s=0#Start index
    i=1#Length of proposed additional match
    while s+i<out_max and a[s:s+i]==b[s:s+i]:
        s+=i
        i*=2
    while i:
        if a[s:s+i]==b[s:s+i]:
            s+=i
        i//=2
    assert a[:s]==b[:s]
    return a[:s]



class LazyLexer:
    """
    Lazy wrapper for Pygments lexers to avoid slow initialization and imports.
    This class was Lazy lexers written by Claude, Aug 5 2025. Resulted in massive pterm boot speed boost. Pygments initializations were taking the majority of the time.
    """
    def __init__(self, lexer_module_path, lexer_class_name, *args, **kwargs):
        self.lexer_module_path = lexer_module_path
        self.lexer_class_name = lexer_class_name
        self.args = args
        self.kwargs = kwargs
        self._lexer = None
        self._lexer_class = None
    
    def _get_lexer_class(self):
        if self._lexer_class is None:
            import importlib
            module = importlib.import_module(self.lexer_module_path)
            self._lexer_class = getattr(module, self.lexer_class_name)
        return self._lexer_class
    
    def _get_lexer(self):
        if self._lexer is None:
            lexer_class = self._get_lexer_class()
            self._lexer = lexer_class(*self.args, **self.kwargs)
        return self._lexer
    
    def get_tokens_unprocessed(self, text):
        return self._get_lexer().get_tokens_unprocessed(text)
    
    def __getattr__(self, name):
        return getattr(self._get_lexer(), name)


class LazyBashLexer(LazyLexer):
    """
    Special lazy wrapper for BashLexer that adds system commands on initialization.
    This class was Lazy lexers written by Claude, Aug 5 2025. Resulted in massive pterm boot speed boost. Pygments initializations were taking the majority of the time.
    """
    _system_commands_added = False  # Class variable to track if we've already modified BashLexer globally
    
    def _get_lexer(self):
        if self._lexer is None:
            # Add system commands to global BashLexer class if not already done
            if not LazyBashLexer._system_commands_added:
                self._add_system_commands_to_pterm_bash_highlighter()
                LazyBashLexer._system_commands_added = True
            
            # Now initialize the lexer (which will inherit the modified tokens)
            lexer_class = self._get_lexer_class()
            self._lexer = lexer_class(*self.args, **self.kwargs)
        return self._lexer
    
    def _add_system_commands_to_pterm_bash_highlighter(self):
        """ 
        This function lets us syntax-highlight any system commands in the !<shell stuff> in pterm seen upon boot 
        It can't update them over time right now, it's a one-time thing
        """ 
        import pygments.lexers.shell as shell
        import re
        import rp
        Name = shell.Name
        commands = rp.get_system_commands() + ['!']
        shell.BashLexer.tokens['basic'] += [(r'(^|!|\b)(' + '|'.join(re.escape(x) for x in commands) + r')(?=[\s)\`]|$)', Name.Function),]


class FastPygmentsTokenizer:
    def __init__(self,pygments_lexer=None):
        if pygments_lexer is None:
            # Lazy import of default lexer
            import six
            if six.PY2:
                from pygments.lexers import PythonLexer
                pygments_lexer = PythonLexer()
            else:
                from pygments.lexers import Python3Lexer
                pygments_lexer = Python3Lexer()
        self.old_text=''
        self.token_cache=[]
        self.pygments_lexer=pygments_lexer
        
        # <CLAUDE CODE START: Initialize language lexers for multi-language highlighting>
        # Common lexer options
        lexer_options = {'stripnl': False, 'stripall': False, 'ensurenl': False}

        # Create lazy lexers - imports and initialization deferred until first use
        self.bash_lexer = LazyBashLexer('pygments.lexers', 'BashLexer', **lexer_options)
        self.javascript_lexer = LazyLexer('pygments.lexers', 'JavascriptLexer', **lexer_options)
        self.html_lexer = LazyLexer('pygments.lexers', 'HtmlLexer', **lexer_options)
        self.sql_lexer = LazyLexer('pygments.lexers', 'SqlLexer', **lexer_options)
        self.json_lexer = LazyLexer('pygments.lexers', 'JsonLexer', **lexer_options)
        self.css_lexer = LazyLexer('pygments.lexers', 'CssLexer', **lexer_options)
        self.xml_lexer = LazyLexer('pygments.lexers', 'XmlLexer', **lexer_options)
        self.yaml_lexer = LazyLexer('pygments.lexers', 'YamlLexer', **lexer_options)
        self.ruby_lexer = LazyLexer('pygments.lexers', 'RubyLexer', **lexer_options)
        self.php_lexer = LazyLexer('pygments.lexers', 'PhpLexer', **lexer_options)
        self.cpp_lexer = LazyLexer('pygments.lexers', 'CppLexer', **lexer_options)
        self.c_lexer = LazyLexer('pygments.lexers', 'CLexer', **lexer_options)
        self.rust_lexer = LazyLexer('pygments.lexers', 'RustLexer', **lexer_options)
        self.go_lexer = LazyLexer('pygments.lexers', 'GoLexer', **lexer_options)
        self.csharp_lexer = LazyLexer('pygments.lexers', 'CSharpLexer', **lexer_options)
        self.java_lexer = LazyLexer('pygments.lexers', 'JavaLexer', **lexer_options)
        self.makefile_lexer = LazyLexer('pygments.lexers', 'MakefileLexer', **lexer_options)
        self.perl_lexer = LazyLexer('pygments.lexers', 'PerlLexer', **lexer_options)
        self.typescript_lexer = LazyLexer('pygments.lexers', 'TypeScriptLexer', **lexer_options)
        self.swift_lexer = LazyLexer('pygments.lexers', 'SwiftLexer', **lexer_options)
        self.kotlin_lexer = LazyLexer('pygments.lexers', 'KotlinLexer', **lexer_options)
        self.scala_lexer = LazyLexer('pygments.lexers', 'ScalaLexer', **lexer_options)
        self.objc_lexer = LazyLexer('pygments.lexers', 'ObjectiveCLexer', **lexer_options)
        self.dart_lexer = LazyLexer('pygments.lexers', 'DartLexer', **lexer_options)
        self.markdown_lexer = LazyLexer('pygments.lexers', 'MarkdownLexer', **lexer_options)
        self.diff_lexer = LazyLexer('pygments.lexers', 'DiffLexer', **lexer_options)
        
        # Map language identifiers to lexers
        self.language_lexers = {
            # Shell scripting
            'bash': self.bash_lexer,
            'sh': self.bash_lexer,
            'shell': self.bash_lexer,
            'zsh': self.bash_lexer,
            
            # Web development
            'javascript': self.javascript_lexer,
            'js': self.javascript_lexer,
            'html': self.html_lexer,
            'css': self.css_lexer,
            'xml': self.xml_lexer,
            'yaml': self.yaml_lexer,
            'yml': self.yaml_lexer,
            'json': self.json_lexer,
            'typescript': self.typescript_lexer,
            'ts': self.typescript_lexer,
            
            # Databases
            'sql': self.sql_lexer,
            
            # Backend languages
            'python': self.pygments_lexer,
            'py': self.pygments_lexer,
            'ruby': self.ruby_lexer,
            'rb': self.ruby_lexer,
            'php': self.php_lexer,
            'perl': self.perl_lexer,
            'pl': self.perl_lexer,
            'java': self.java_lexer,
            'scala': self.scala_lexer,
            'kotlin': self.kotlin_lexer,
            'kt': self.kotlin_lexer,
            
            # Systems programming
            'c': self.c_lexer,
            'cpp': self.cpp_lexer,
            'c++': self.cpp_lexer,
            'cxx': self.cpp_lexer,
            'rust': self.rust_lexer,
            'rs': self.rust_lexer,
            'go': self.go_lexer,
            'golang': self.go_lexer,
            'csharp': self.csharp_lexer,
            'cs': self.csharp_lexer,
            
            # Mobile development
            'swift': self.swift_lexer,
            'objc': self.objc_lexer,
            'objectivec': self.objc_lexer,
            'dart': self.dart_lexer,
            
            # Other useful formats
            'makefile': self.makefile_lexer,
            'make': self.makefile_lexer,
            'markdown': self.markdown_lexer,
            'md': self.markdown_lexer,
            'diff': self.diff_lexer,
            'patch': self.diff_lexer,
        }
        # <CLAUDE CODE END: Initialize language lexers for multi-language highlighting>
    def _set_new_text(self,text):
        # from rp import longest_common_prefix
        #Used to invalidate the token_cache
        prefix=longest_common_prefix(text,self.old_text)
        length=len(prefix)
        # self.token_cache=[token for token in self.token_cache if token[0]<length]
        from bisect import bisect_left
        try:
            del self.token_cache[max(0,bisect_left(self.token_cache,(length,))-1):]
        except:
            self.token_cache=[]
        try:
            while self.token_cache[-1][1]!='\n':
            # for _ in range(10):
                self.token_cache.pop()
            #Snip off the tail just to be sure its correct, in particular because '"HELLO"' is treated as 3 tokens ('"','HELLO','"') we need to make sure the beginning quote is kept when re-highlighting
        except:
            pass
        self.old_text=text
    def get_tokens_unprocessed(self,text):

        if text.startswith('!'):
            #ORIGINAL CODE: Not for nested languages! 

            # yield (0, Token.Keyword, text)
            #Yields something like [(0, Token.Keyword, 'def'), (3, Token.Text, ' '), (4, Token.Name.Function, 'f'), (5, Token.Punctuation, '('), (6, Token.Punctuation, ')'), (7, Token.Punctuation, ':'), (8, Token.Keyword, 'pass')]
            self._set_new_text(text)
            # from rp import text_to_speech
            # text_to_speech(len(self.token_cache))
            yield from self.token_cache
            # for token in self.token_cache:
                # yield token
            if not self.token_cache:
                start_pos=0
            else:
                start_pos=self.token_cache[-1][0]+len(self.token_cache[-1][2])
            if start_pos>=len(text):
                return#We're aleady at the end of the string; we're done. no more tokens.
            else:
                for token in self.pygments_lexer.get_tokens_unprocessed(text[start_pos:]):
                    token=(token[0]+start_pos,token[1],token[2])
                    # start,species,data=token
                    # start+=start_pos
                    # token=start,species,data
                    self.token_cache.append(token)
                    yield token
            return


        # yield (0, Token.Keyword, text)
        #Yields something like [(0, Token.Keyword, 'def'), (3, Token.Text, ' '), (4, Token.Name.Function, 'f'), (5, Token.Punctuation, '('), (6, Token.Punctuation, ')'), (7, Token.Punctuation, ':'), (8, Token.Keyword, 'pass')]
        self._set_new_text(text)
        # from rp import text_to_speech
        # text_to_speech(len(self.token_cache))
        yield from self.token_cache
        # for token in self.token_cache:
            # yield token
        if not self.token_cache:
            start_pos=0
        else:
            start_pos=self.token_cache[-1][0]+len(self.token_cache[-1][2])
        if start_pos>=len(text):
            return#We're aleady at the end of the string; we're done. no more tokens.
        else:
            # <CLAUDE CODE START: Improved Multi-language string highlighting implementation>
            # Get the raw tokens from the Pygments lexer
            raw_tokens = list(self.pygments_lexer.get_tokens_unprocessed(text[start_pos:]))
            
            # Process tokens for special language highlighting
            i = 0
            while i < len(raw_tokens):
                pos, token_type, token_text = raw_tokens[i]
                adjusted_pos = pos + start_pos
                
                # Define string token types to look for
                string_token_types = [
                    Token.Literal.String,
                    Token.Literal.String.Doc,
                    Token.Literal.String.Double,
                    Token.Literal.String.Single,
                    Token.Literal.String.Backtick,
                    Token.Literal.String.Heredoc,
                ]
                
                # Check if this is a string token
                if any(token_type == t or str(token_type).startswith(str(t) + '.') for t in string_token_types):
                    # This is the start of a string - collect all consecutive string tokens
                    string_tokens = [(pos, token_type, token_text)]
                    string_start = i
                    current_pos = pos + len(token_text)
                    j = i + 1
                    
                    # Collect all consecutive string tokens
                    while j < len(raw_tokens):
                        next_pos, next_type, next_text = raw_tokens[j]
                        # If not consecutive or not a string token, break
                        if next_pos != current_pos or not any(next_type == t or str(next_type).startswith(str(t) + '.') for t in string_token_types):
                            break
                        
                        # Add this string token to our collection
                        string_tokens.append((next_pos, next_type, next_text))
                        current_pos = next_pos + len(next_text)
                        j += 1
                    
                    # If we have more than one string token or a single one with a language tag, process it
                    if len(string_tokens) > 0 or '#!' in token_text:
                        # Combine all string tokens into a single string
                        combined_text = ''.join(t[2] for t in string_tokens)
                        combined_start = string_tokens[0][0]
                        combined_pos = combined_start + start_pos
                        
                        # Detect quote type and extract content
                        opening_quotes = ""
                        closing_quotes = ""
                        content = combined_text
                        
                        # Identify opening quotes, including prefix modifiers (r, f, etc.)
                        # Regular expressions would be better here, but we'll use simple string checks
                        prefix = ""
                        
                        # Check for r-strings, f-strings, etc.
                        if any(combined_text.lower().startswith(p) for p in ['r', 'f', 'b', 'u', 'fr', 'rf', 'br', 'rb']):
                            # Extract the prefix (r, f, etc.)
                            for p in ['fr', 'rf', 'br', 'rb', 'r', 'f', 'b', 'u']:
                                if combined_text.lower().startswith(p):
                                    prefix_end = len(p)
                                    prefix = combined_text[:prefix_end]
                                    content = combined_text[prefix_end:]
                                    break
                        
                        # Now check for the actual quotes
                        if content.startswith('"""'):
                            opening_quotes = prefix + '"""'
                            content = content[3:]
                        elif content.startswith("'''"):
                            opening_quotes = prefix + "'''"
                            content = content[3:]
                        elif content.startswith('"'):
                            opening_quotes = prefix + '"'
                            content = content[1:]
                        elif content.startswith("'"):
                            opening_quotes = prefix + "'"
                            content = content[1:]
                        elif not prefix and combined_text.startswith('"""'):
                            # No prefix detected but still starts with quotes
                            opening_quotes = '"""'
                            content = combined_text[3:]
                        elif not prefix and combined_text.startswith("'''"):
                            opening_quotes = "'''"
                            content = combined_text[3:]
                        elif not prefix and combined_text.startswith('"'):
                            opening_quotes = '"'
                            content = combined_text[1:]
                        elif not prefix and combined_text.startswith("'"):
                            opening_quotes = "'"
                            content = combined_text[1:]
                        else:
                            # Failed to detect opening quotes
                            opening_quotes = ""
                            content = combined_text
                        
                        # Identify closing quotes
                        if content.endswith('"""'):
                            closing_quotes = '"""'
                            content = content[:-3]
                        elif content.endswith("'''"):
                            closing_quotes = "'''"
                            content = content[:-3]
                        elif content.endswith('"'):
                            closing_quotes = '"'
                            content = content[:-1]
                        elif content.endswith("'"):
                            closing_quotes = "'"
                            content = content[:-1]
                        
                        # Check for language tag in the content
                        bang_pos = combined_text.find('#!')
                        language_found = False
                        
                        if bang_pos >= 0:
                            # Extract language identifier
                            first_line_end = combined_text.find('\n', bang_pos)
                            if first_line_end == -1:
                                first_line_end = len(combined_text)
                            
                            lang_tag = combined_text[bang_pos+2:first_line_end].strip()
                            lang = lang_tag.split()[0].split('/')[-1] if '/' in lang_tag else lang_tag
                            
                            # If we recognize this language, use its lexer
                            if lang.lower() in self.language_lexers:
                                language_found = True
                                
                                # Process using language-specific lexer
                                tokens_to_yield = []
                                
                                # Calculate the positions of different parts
                                if opening_quotes:
                                    # 1. Opening quotes as string literal
                                    tokens_to_yield.append(
                                        (combined_pos, Token.Literal.String, opening_quotes)
                                    )
                                    
                                    # 2. Content before the language tag, if any
                                    if bang_pos > len(opening_quotes):
                                        prefix = combined_text[len(opening_quotes):bang_pos]
                                        tokens_to_yield.append(
                                            (combined_pos + len(opening_quotes), Token.Text, prefix)
                                        )
                                    
                                    # 3. The language tag as a hashbang
                                    shebang_text = combined_text[bang_pos:first_line_end].strip()
                                    tokens_to_yield.append(
                                        (combined_pos + bang_pos, Token.Comment.Hashbang, shebang_text)
                                    )
                                    
                                    # 4. Add the newline if present
                                    if first_line_end < len(combined_text):
                                        tokens_to_yield.append(
                                            (combined_pos + first_line_end, Token.Text, '\n')
                                        )
                                        
                                        # 5. Content after the first line with language-specific highlighting
                                        remaining_content = combined_text[first_line_end+1:]
                                        if closing_quotes and remaining_content.endswith(closing_quotes):
                                            remaining_content = remaining_content[:-len(closing_quotes)]
                                        
                                        # Use language-specific lexer for the main content
                                        language_lexer = self.language_lexers[lang.lower()]
                                        content_tokens = language_lexer.get_tokens_unprocessed(remaining_content)
                                        
                                        # Add with adjusted positions
                                        content_start = combined_pos + first_line_end + 1
                                        for content_pos, content_type, content_text in content_tokens:
                                            tokens_to_yield.append(
                                                (content_start + content_pos, content_type, content_text)
                                            )
                                    
                                    # 6. Add closing quotes if present
                                    if closing_quotes:
                                        closing_pos = combined_pos + len(combined_text) - len(closing_quotes)
                                        tokens_to_yield.append(
                                            (closing_pos, Token.Literal.String, closing_quotes)
                                        )
                                
                                # Yield all tokens
                                for token in tokens_to_yield:
                                    self.token_cache.append(token)
                                    yield token
                                
                                # Skip all the tokens we've processed
                                i = j
                                continue
                        
                        # Check for trailing comment-style language tag
                        elif j < len(raw_tokens) and raw_tokens[j][1] == Token.Comment.Single:
                            comment_pos, comment_type, comment_text = raw_tokens[j]
                            if comment_text.startswith('#'):
                                lang = comment_text[1:].strip()
                                if lang.lower() in self.language_lexers:
                                    language_found = True
                                    
                                    # Process using language-specific lexer with comment tag
                                    tokens_to_yield = []
                                    
                                    # 1. Opening quotes
                                    if opening_quotes:
                                        tokens_to_yield.append(
                                            (combined_pos, Token.Literal.String, opening_quotes)
                                        )
                                    
                                    # 2. Content with language-specific highlighting
                                    language_lexer = self.language_lexers[lang.lower()]
                                    content_tokens = language_lexer.get_tokens_unprocessed(content)
                                    
                                    # Add with adjusted positions
                                    content_start = combined_pos + len(opening_quotes)
                                    for content_pos, content_type, content_text in content_tokens:
                                        tokens_to_yield.append(
                                            (content_start + content_pos, content_type, content_text)
                                        )
                                    
                                    # 3. Closing quotes
                                    if closing_quotes:
                                        closing_pos = combined_pos + len(combined_text) - len(closing_quotes)
                                        tokens_to_yield.append(
                                            (closing_pos, Token.Literal.String, closing_quotes)
                                        )
                                    
                                    # 4. The language tag comment
                                    tokens_to_yield.append(
                                        (comment_pos + start_pos, Token.Comment.Single, comment_text)
                                    )
                                    
                                    # Yield all tokens
                                    for token in tokens_to_yield:
                                        self.token_cache.append(token)
                                        yield token
                                    
                                    # Skip all the tokens we've processed including the comment
                                    i = j + 1
                                    continue
                    
                    # If no language tag was found, or language not supported,
                    # just yield all collected string tokens normally
                    if not language_found:
                        for str_pos, str_type, str_text in string_tokens:
                            adjusted_token = (str_pos + start_pos, str_type, str_text)
                            self.token_cache.append(adjusted_token)
                            yield adjusted_token
                        i = j
                        continue
                
                # Regular token processing for non-string tokens
                adjusted_token = (pos + start_pos, token_type, token_text)
                self.token_cache.append(adjusted_token)
                yield adjusted_token
                i += 1
            # <CLAUDE CODE END: Improved Multi-language string highlighting implementation>
# t=FastPygmentsTokenizer()
# print(list(t.get_tokens_unprocessed('((HELLO))')))
# print(list(t.get_tokens_unprocessed('((HELLO)))')))
# print(list(t.get_tokens_unprocessed('((HO)))')))

        
        
        
        

class PygmentsLexer(Lexer):
    """
    Lexer that calls a pygments lexer.

    Example::

        from pygments.lexers import HtmlLexer
        lexer = PygmentsLexer(HtmlLexer)

    Note: Don't forget to also load a Pygments compatible style. E.g.::

        from rp.prompt_toolkit.styles.from_pygments import style_from_pygments
        from pygments.styles import get_style_by_name
        style = style_from_pygments(get_style_by_name('monokai'))

    :param pygments_lexer_cls: A `Lexer` from Pygments.
    :param sync_from_start: Start lexing at the start of the document. This
        will always give the best results, but it will be slow for bigger
        documents. (When the last part of the document is display, then the
        whole document will be lexed by Pygments on every key stroke.) It is
        recommended to disable this for inputs that are expected to be more
        than 1,000 lines.
    :param syntax_sync: `SyntaxSync` object.
    """
    # Minimum amount of lines to go backwards when starting the parser.
    # This is important when the lines are retrieved in reverse order, or when
    # scrolling upwards. (Due to the complexity of calculating the vertical
    # scroll offset in the `Window` class, lines are not always retrieved in
    # order.)
    MIN_LINES_BACKWARDS = 50

    # When a parser was started this amount of lines back, read the parser
    # until we get the current line. Otherwise, start a new parser.
    # (This should probably be bigger than MIN_LINES_BACKWARDS.)
    REUSE_GENERATOR_MAX_DISTANCE = 100

    def __init__(self, pygments_lexer_cls, sync_from_start=True, syntax_sync=None):
        assert syntax_sync is None or isinstance(syntax_sync, SyntaxSync)

        self.pygments_lexer_cls = pygments_lexer_cls
        self.sync_from_start = to_cli_filter(sync_from_start)

        self.old_text=""

        # Instantiate the Pygments lexer. (Python)
        self.pygments_lexer = pygments_lexer_cls(
            stripnl=False,
            stripall=False,
            ensurenl=False,
        )
        self.fast_pygments_lexer=self.pygments_lexer
        self.fast_pygments_tokenizer=FastPygmentsTokenizer(self.pygments_lexer)

        # Create lazy Bash lexer and tokenizer - only initialized when needed
        self._lazy_bash_lexer = None
        self._lazy_bash_tokenizer = None

        # Create syntax sync instance.
        self.syntax_sync = syntax_sync or RegexSync.from_pygments_lexer_cls(pygments_lexer_cls)

    @property
    def fast_bash_lexer(self):
        if self._lazy_bash_lexer is None:
            self._lazy_bash_lexer = LazyBashLexer(
                'pygments.lexers',
                'BashLexer',
                stripnl=False,
                stripall=False,
                ensurenl=False,
            )
        return self._lazy_bash_lexer

    @property  
    def fast_bash_tokenizer(self):
        if self._lazy_bash_tokenizer is None:
            self._lazy_bash_tokenizer = FastPygmentsTokenizer(self.fast_bash_lexer)
        return self._lazy_bash_tokenizer

    def get_lexer(self, document):
        text = document.text
        if text.startswith("!"):
            return self.fast_bash_lexer  # Bash
        else:
            return self.fast_pygments_lexer  # Python

    # <CLAUDE CODE START: Update tokenizer selection to check for language tags>
    def get_tokenizer(self, document):
        text = document.text
        if text.startswith("!"):
            return self.fast_bash_tokenizer  # Bash
        # Check for multiline string with language shebang
        elif '"""#!' in text or "'''#!" in text or ('#!' in text and ('"""' in text or "'''" in text)):
            # Return Python lexer for now, we'll handle language-specific highlights later
            return self.fast_pygments_tokenizer  # Python
        else:
            return self.fast_pygments_tokenizer  # Python
    # <CLAUDE CODE END: Update tokenizer selection to check for language tags>

    @classmethod
    def from_filename(cls, filename, sync_from_start=True):
        """
        Create a `Lexer` from a filename.
        """
        # Inline imports: the Pygments dependency is optional!
        from pygments.util import ClassNotFound
        from pygments.lexers import get_lexer_for_filename

        try:
            pygments_lexer = get_lexer_for_filename(filename)
        except ClassNotFound:
            return SimpleLexer()
        else:
            return cls(pygments_lexer.__class__, sync_from_start=sync_from_start)

    def lex_document(self, cli, document):
        """
        Create a lexer function that takes a line number and returns the list
        of (Token, text) tuples as the Pygments lexer returns for that line.
        """
        # Cache of already lexed lines.
        cache = {}

        # Pygments generators that are currently lexing.
        line_generators = {}  # Map lexer generator to the line number.

        def get_syntax_sync():
            " The Syntax synchronisation objcet that we currently use. "
            if self.sync_from_start(cli):
                return SyncFromStart()
            else:
                return self.syntax_sync

        def find_closest_generator(i):
            " Return a generator close to line 'i', or None if none was fonud. "
            for generator, lineno in line_generators.items():
                if lineno < i and i - lineno < self.REUSE_GENERATOR_MAX_DISTANCE:
                    return generator

        def create_line_generator(start_lineno, column=0):
            """
            Create a generator that yields the lexed lines.
            Each iteration it yields a (line_number, [(token, text), ...]) tuple.
            """
            def get_tokens():
                text = '\n'.join(document.lines[start_lineno:])[column:]
                text = '\n'.join(document.lines[0:])[0:]

                # We call `get_tokens_unprocessed`, because `get_tokens` will
                # still replace \r\n and \r by \n.  (We don't want that,
                # Pygments should return exactly the same amount of text, as we
                # have given as input.)
                for _, t, v in self.get_tokenizer(document).get_tokens_unprocessed(text):
                    yield t, v

            return enumerate(split_lines(get_tokens()), start_lineno)

        def get_generator(i):
            """
            Find an already started generator that is close, or create a new one.
            """
            # Find closest line generator.
            generator = find_closest_generator(i)
            if generator:
                return generator

            # No generator found. Determine starting point for the syntax
            # synchronisation first.

            # Go at least x lines back. (Make scrolling upwards more
            # efficient.)
            i = max(0, i - self.MIN_LINES_BACKWARDS)

            if i == 0:
                row = 0
                column = 0
            else:
                row, column = get_syntax_sync().get_sync_start_position(document, i)

            # Find generator close to this point, or otherwise create a new one.
            generator = find_closest_generator(i)
            if generator:
                return generator
            else:
                generator = create_line_generator(row, column)

            # If the column is not 0, ignore the first line. (Which is
            # incomplete. This happens when the synchronisation algorithm tells
            # us to start parsing in the middle of a line.)
            if column:
                next(generator)
                row += 1

            line_generators[generator] = row
            return generator

        def get_line(i):
            " Return the tokens for a given line number. "
            try:
                return cache[i]
            except KeyError:
                generator = get_generator(i)

                # Exhaust the generator, until we find the requested line.
                for num, line in generator:
                    cache[num] = line
                    if num == i:
                        line_generators[generator] = i

                        # Remove the next item from the cache.
                        # (It could happen that it's already there, because of
                        # another generator that started filling these lines,
                        # but we want to synchronise these lines with the
                        # current lexer's state.)
                        if num + 1 in cache:
                            del cache[num + 1]

                        return cache[num]
            return []
        self.old_text=document.text#For speed's sake: Ryan Burgert
        return get_line


class LazyPygmentsLexer(Lexer):
    """
    Lazy wrapper for PygmentsLexer that defers lexer imports until first use.
    This class was Written by Claude, Aug 5 2025. Resulted in massive pterm boot speed boost. Pygments initializations were taking the majority of the time.
    """
    
    def __init__(self):
        self._real_lexer = None
    
    def _get_real_lexer(self):
        if self._real_lexer is None:
            # Import only when needed
            import six
            if six.PY2:
                from pygments.lexers import PythonLexer
                lexer_class = PythonLexer
            else:
                from pygments.lexers import Python3Lexer as PythonLexer
                lexer_class = PythonLexer
            
            # Create the actual PygmentsLexer
            self._real_lexer = PygmentsLexer(lexer_class)
        return self._real_lexer
    
    def lex_document(self, cli, document):
        return self._get_real_lexer().lex_document(cli, document)
    
    def __getattr__(self, name):
        return getattr(self._get_real_lexer(), name)



from jsonpath_ng import parse
from loguru import logger

from .autocomplete import autocomplete
from .io.datasource import create_data_source


class Model:
    """
    pyfx model entry point, which loads and processes JSON data.

    Currently it manages the following actions:
     * loads the original JSON file into memory
     * parses JSONPath query and returns new data
     * performs auto-completion with given JSONPath query
    """

    def __init__(self, dispatcher):
        self._data = None
        self._current = None
        dispatcher.register("load", self.load)
        dispatcher.register("query", self.query)
        dispatcher.register("complete", self.complete)

    def load(self, type, *args):
        datasource = create_data_source(type)(*args)
        self._data = datasource.read_json()
        self._current = self._data
        return self._current

    def query(self, text):
        if self._data is None:
            logger.debug("Data is None.")
            return None

        result = self._query(text)
        self._current = result[0] if len(result) == 1 else result
        return self._current

    def complete(self, text):
        if self._data is None:
            logger.debug("Data is None.")
            return False, "", []
        return autocomplete(text, self.query)

    # noinspection PyBroadException
    def _query(self, text):
        try:
            jsonpath_expr = parse(text)
            return [match.value for match in jsonpath_expr.find(self._data)]
        except Exception as e:
            logger.opt(exception=True) \
                .error("JSONPath query '{}' failed with: {}", text, e)
            return []



from overrides import overrides

from ..json_widget import JSONWidget


class ObjectEndWidget(JSONWidget):
    """
    a widget to display the end symbol for JSON `object` type nodes
    """

    def __init__(self, node):
        super().__init__(node, True, False)

    @overrides
    def load_value_markup(self):
        return "}"



from typing import Any
from werkzeug.wrappers import Response

class ContentAccessors:
    def xml(self): ...
    def lxml(self): ...
    def json(self): ...

class TestResponse(Response, ContentAccessors): ...



import pathlib

import dacite
import yamale
from first import first
from yamale import YamaleError
from yamale.validators import DefaultValidators

from .config_error import ConfigurationError
from .validators import Options
from .config import Configuration


def parse(config_file=None):
    try:
        return ConfigurationParser().parse(config_file)
    except YamaleError as e:
        # catch and raise a more user-friendly error
        message = '\n'.join(e.message.split('\n')[1:]).strip()
        raise ConfigurationError(message)


class ConfigurationParser:

    __CLASS_DIR = pathlib.Path(__file__).parent.resolve()
    __SCHEMA_PATH = __CLASS_DIR / "schema.yml"

    __CONFIG_PATHS = [
        pathlib.Path.home() / ".config" / "pyfx" / "config.yml",
        __CLASS_DIR / "config.yml"
    ]

    def __init__(self):
        self._schema = self.__load_schema()

    def parse(self, config_file=None):
        config = self.__load_config(config_file)
        yamale.validate(self._schema, config)
        # config is composed as [(config, path)...]
        return dacite.from_dict(data_class=Configuration, data=config[0][0])

    @staticmethod
    def __load_schema():
        validators = DefaultValidators.copy()
        validators[Options.tag] = Options
        return yamale.make_schema(
            ConfigurationParser.__SCHEMA_PATH, validators=validators
        )

    # noinspection PyBroadException
    @staticmethod
    def __load_config(config_file):
        if config_file is None:
            config_file = first(
                ConfigurationParser.__CONFIG_PATHS,
                key=lambda path: path.exists()
            )
        return yamale.make_data(config_file)
